from __future__ import print_function
import numpy as np
import os
from gym import spaces

from navrep.tools.rings import generate_rings
from navrep.envs.ianenv import IANEnv
from navrep.models.rnn import (reset_graph, sample_hps_params, MDNRNN,
                               rnn_init_state, rnn_next_state, MAX_GOAL_DIST)
from navrep.models.vae2d import ConvVAE
from navrep.models.vae1d import Conv1DVAE
from navrep.models.gpt import GPT, GPTConfig, load_checkpoint
from navrep.models.gpt1d import GPT1D
from navrep.models.vae1dlstm import VAE1DLSTM, VAE1DLSTMConfig
from navrep.models.vaelstm import VAELSTM, VAELSTMConfig
from navrep.tools.wdataset import scans_to_lidar_obs

PUNISH_SPIN = True

""" VM backends: VAE_LSTM, W backends: GPT, GPT1D, VAE1DLSTM """
""" ENCODINGS: V_ONLY, VM, M_ONLY """
_G = 2  # goal dimensions
_A = 3  # action dimensions
_RS = 5  # robot state
_64 = 64  # ring size
_L = 1080  # lidar size
NO_VAE_VAR = True

BLOCK_SIZE = 32  # sequence length (context)

class EnvEncoder(object):
    """ Generic class to encode the observations of an environment,
    look at EncodedEnv to see how it is typically used """
    def __init__(self,
                 backend, encoding,
                 rnn_model_path=os.path.expanduser("~/navrep/models/M/rnn.json"),
                 rnn1d_model_path=os.path.expanduser("~/navrep/models/M/rnn1d.json"),
                 vae_model_path=os.path.expanduser("~/navrep/models/V/vae.json"),
                 vae1d_model_path=os.path.expanduser("~/navrep/models/V/vae1d.json"),
                 gpt_model_path=os.path.expanduser("~/navrep/models/W/gpt"),
                 gpt1d_model_path=os.path.expanduser("~/navrep/models/W/gpt1d"),
                 vae1dlstm_model_path=os.path.expanduser("~/navrep/models/W/vae1dlstm"),
                 vaelstm_model_path=os.path.expanduser("~/navrep/models/W/vaelstm"),
                 gpu=False,
                 encoder_to_share_model_with=None,  # another EnvEncoder
                 ):
        LIDAR_NORM_FACTOR = None
        if backend == "GPT":
            from navrep.scripts.train_gpt import _Z, _H
        elif backend == "GPT1D":
            from navrep.scripts.train_gpt1d import _Z, _H
            from navrep.tools.wdataset import LIDAR_NORM_FACTOR
        elif backend == "VAE1DLSTM":
            from navrep.scripts.train_vae1dlstm import _Z, _H
            from navrep.tools.wdataset import LIDAR_NORM_FACTOR
        elif backend == "VAELSTM":
            from navrep.scripts.train_vaelstm import _Z, _H
        elif backend == "VAE_LSTM":
            from navrep.scripts.train_vae import _Z
            from navrep.scripts.train_rnn import _H
        elif backend == "VAE1D_LSTM":
            from navrep.scripts.train_vae1d import _Z
            from navrep.scripts.train_rnn import _H
            from navrep.scripts.train_vae1d import MAX_LIDAR_DIST as LIDAR_NORM_FACTOR
        self._Z = _Z
        self._H = _H
        self.LIDAR_NORM_FACTOR = LIDAR_NORM_FACTOR
        self.encoding = encoding
        self.backend = backend
        if self.encoding == "V_ONLY":
            self.encoding_dim = _Z + _RS
        elif self.encoding == "VM":
            self.encoding_dim = _Z + _H + _RS
        elif self.encoding == "M_ONLY":
            self.encoding_dim = _H + _RS
        else:
            raise NotImplementedError
        self.observation_space = spaces.Box(low=-np.inf, high=np.inf,
                                            shape=(self.encoding_dim,), dtype=np.float32)
        # V + M Models
        if encoder_to_share_model_with is not None:
            self.vae = encoder_to_share_model_with.vae
            self.rnn = encoder_to_share_model_with.rnn
        else:
            # load world model
            if self.backend == "VAE_LSTM":
                reset_graph()
                self.vae = ConvVAE(z_size=_Z, batch_size=1, is_training=False)
                self.vae.load_json(vae_model_path)
                if self.encoding in ["VM", "M_ONLY"]:
                    hps = sample_hps_params. _replace(seq_width=_Z+_G, action_width=_A, rnn_size=_H)
                    self.rnn = MDNRNN(hps, gpu_mode=gpu)
                    self.rnn.load_json(rnn_model_path)
            elif self.backend == "VAE1D_LSTM":
                reset_graph()
                self.vae = Conv1DVAE(z_size=_Z, batch_size=1, is_training=False)
                self.vae.load_json(vae1d_model_path)
                if self.encoding in ["VM", "M_ONLY"]:
                    hps = sample_hps_params. _replace(seq_width=_Z+_G, action_width=_A, rnn_size=_H)
                    self.rnn = MDNRNN(hps, gpu_mode=gpu)
                    self.rnn.load_json(rnn1d_model_path)
            elif self.backend == "GPT":
                mconf = GPTConfig(BLOCK_SIZE, _H)
                model = GPT(mconf, gpu=gpu)
                load_checkpoint(model, gpt_model_path, gpu=gpu)
                self.vae = model
                self.rnn = model
            elif self.backend == "GPT1D":
                mconf = GPTConfig(BLOCK_SIZE, _H)
                model = GPT1D(mconf, gpu=gpu)
                load_checkpoint(model, gpt1d_model_path, gpu=gpu)
                self.vae = model
                self.rnn = model
            elif self.backend == "VAELSTM":
                mconf = VAELSTMConfig(_Z, _H)
                model = VAELSTM(mconf, gpu=gpu)
                load_checkpoint(model, vaelstm_model_path, gpu=gpu)
                self.vae = model
                self.rnn = model
            elif self.backend == "VAE1DLSTM":
                mconf = VAE1DLSTMConfig(_Z, _H)
                model = VAE1DLSTM(mconf, gpu=gpu)
                load_checkpoint(model, vae1dlstm_model_path, gpu=gpu)
                self.vae = model
                self.rnn = model
            else:
                raise NotImplementedError
        # other tools
        self.rings_def = generate_rings(_64, _64)
        self.viewer = None
        # environment state variables
        self.reset()

    def reset(self):
        if self.encoding in ["VM", "M_ONLY"]:
            if self.backend in ["VAE_LSTM", "VAE1D_LSTM"]:
                self.state = rnn_init_state(self.rnn)
            elif self.backend in ["GPT", "VAELSTM", "VAE1DLSTM", "GPT1D"]:
                self.gpt_sequence = []
        self.lidar_z = np.zeros(self._Z)

    def close(self):
        if self.viewer is not None:
            self.viewer.close()

    def _get_last_decoded_scan(self):
        obs_pred = self.vae.decode(self.lidar_z.reshape((1,self._Z)))
        if self.backend in ["VAE1DLSTM", "GPT1D", "VAE1D_LSTM"]:
            decoded_scan = (obs_pred * self.LIDAR_NORM_FACTOR).reshape((_L))
        else:
            rings_pred = obs_pred * self.rings_def["rings_to_bool"]
            decoded_scan = self.rings_def["rings_to_lidar"](rings_pred, _L).reshape((_L))
        return decoded_scan

    def _encode_obs(self, obs, action):
        """
    obs is (lidar, other_obs)
    where lidar is (time_samples, ray, channel)
    and other_obs is (5,) - [goal_x, goal_y, vel_x, vel_y, vel_theta] all in robot frame

    h is (32+2+512), i.e. concat[lidar_z, robotstate, h rnn state]
    lidar_z is -inf, inf
    h rnn state is ?
    other_obs is -inf, inf
    """
        # convert lidar scan to obs
        lidar_scan = obs[0]  # latest scan only obs (buffer, ray, channel)
        lidar_scan = lidar_scan.reshape(1, _L).astype(np.float32)
        lidar_mode = "scans" if "1D" in self.backend else "rings"
        lidar_obs = scans_to_lidar_obs(lidar_scan, lidar_mode, self.rings_def, channel_first=False)
        self.last_lidar_obs = lidar_obs  # for rendering purposes

        # obs to z, mu, logvar
        mu, logvar = self.vae.encode_mu_logvar(lidar_obs)
        mu = mu[0]
        logvar = logvar[0]
        s = logvar.shape
        if NO_VAE_VAR:
            lidar_z = mu * 1.
        else:
            lidar_z = mu + np.exp(logvar / 2.0) * np.random.randn(*s)

        # encode obs through V + M
        self.lidar_z = lidar_z
        if self.encoding == "V_ONLY":
            encoded_obs = np.concatenate([self.lidar_z, obs[1]], axis=0)
        elif self.encoding in ["VM", "M_ONLY"]:
            # get h
            if self.backend in ["VAE_LSTM", "VAE1D_LSTM"]:
                goal_z = obs[1][:2] / MAX_GOAL_DIST
                rnn_z = np.concatenate([lidar_z, goal_z], axis=-1)
                self.state = rnn_next_state(self.rnn, rnn_z, action, self.state)
                h = self.state.h[0]
            elif self.backend in ["GPT", "VAELSTM", "VAE1DLSTM", "GPT1D"]:
                self.gpt_sequence.append(dict(obs=lidar_obs[0], state=obs[1][:2], action=action))
                self.gpt_sequence = self.gpt_sequence[:BLOCK_SIZE]
                h = self.rnn.get_h(self.gpt_sequence)
            # encoded obs
            if self.encoding == "VM":
                encoded_obs = np.concatenate([self.lidar_z, obs[1], h], axis=0)
            elif self.encoding == "M_ONLY":
                encoded_obs = np.concatenate([obs[1], h], axis=0)
        return encoded_obs

    def _render_rings_polar(self, close, save_to_file=False):
        if close:
            self.viewer.close()
            return
        # rendering
        if self.backend in ["VAE1DLSTM", "GPT1D", "VAE1D_LSTM"]:
            return False
        else:
            last_rings_obs = self.last_lidar_obs.reshape((_64, _64, 1))
            last_rings_pred = self.vae.decode(self.lidar_z.reshape((1,self._Z))).reshape((_64, _64, 1))
            import matplotlib.pyplot as plt
            plt.ion()
            fig, (ax1, ax2) = plt.subplots(
                1, 2, subplot_kw=dict(projection="polar"), num="rings"
            )
            ax1.clear()
            ax2.clear()
            if self.viewer is None:
                self.rendering_iteration = 0
            self.viewer = fig
            self.rings_def["visualize_rings"](last_rings_obs, scan=None, fig=fig, ax=ax1)
            self.rings_def["visualize_rings"](last_rings_pred, scan=None, fig=fig, ax=ax2)
            ax1.set_ylim([0, 10])
            ax1.set_title("ground truth")
            ax2.set_ylim([0, 10])
            ax2.set_title("lidar reconstruction")
            # rings box viz
            fig2, (ax1, ax2) = plt.subplots(1, 2, num="2d")
            ax1.clear()
            ax2.clear()
            ax1.imshow(np.squeeze(last_rings_obs), cmap=plt.cm.Greys)
            ax2.imshow(np.squeeze(last_rings_pred), cmap=plt.cm.Greys)
            ax1.set_title("ground truth")
            ax2.set_title("lidar reconstruction")
            # update
            plt.pause(0.01)
            self.rendering_iteration += 1
            if save_to_file:
                fig.savefig(
                    "/tmp/encodedenv_polar{:04d}.png".format(self.rendering_iteration))
                fig2.savefig(
                    "/tmp/encodedenv_box{:04d}.png".format(self.rendering_iteration))

    def _render_rings(self, close, save_to_file=False):
        if close:
            self.viewer.close()
            return
        # rendering
        if self.backend in ["VAE1DLSTM", "GPT1D", "VAE1D_LSTM"]:
            return False
        else:
            last_rings_obs = self.last_lidar_obs.reshape((_64, _64))
            last_rings_pred = self.vae.decode(self.lidar_z.reshape((1,self._Z))).reshape((_64, _64))
            # Window and viewport size
            ring_size = _64  # grid cells
            padding = 4  # grid cells
            grid_size = 1  # px per grid cell
            WINDOW_W = (2 * ring_size + 3 * padding) * grid_size
            WINDOW_H = (1 * ring_size + 2 * padding) * grid_size
            VP_W = WINDOW_W
            VP_H = WINDOW_H
            from gym.envs.classic_control import rendering
            import pyglet
            from pyglet import gl
            # Create viewer
            if self.viewer is None:
                self.viewer = rendering.Viewer(WINDOW_W, WINDOW_H)
                self.rendering_iteration = 0
            # Render in pyglet
            win = self.viewer.window
            win.switch_to()
            win.dispatch_events()
            win.clear()
            gl.glViewport(0, 0, VP_W, VP_H)
            # colors
            bgcolor = np.array([0.4, 0.8, 0.4])
            # Green background
            gl.glBegin(gl.GL_QUADS)
            gl.glColor4f(bgcolor[0], bgcolor[1], bgcolor[2], 1.0)
            gl.glVertex3f(0, VP_H, 0)
            gl.glVertex3f(VP_W, VP_H, 0)
            gl.glVertex3f(VP_W, 0, 0)
            gl.glVertex3f(0, 0, 0)
            gl.glEnd()
            # rings - observation
            w_offset = 0
            for rings in [last_rings_obs, last_rings_pred]:
                for i in range(ring_size):
                    for j in range(ring_size):
                        cell_color = 1 - rings[i, j]
                        cell_y = (padding + i) * grid_size  # px
                        cell_x = (padding + j + w_offset) * grid_size  # px
                        gl.glBegin(gl.GL_QUADS)
                        gl.glColor4f(cell_color, cell_color, cell_color, 1.0)
                        gl.glVertex3f(cell_x+       0,  cell_y+grid_size, 0)  # noqa
                        gl.glVertex3f(cell_x+grid_size, cell_y+grid_size, 0)  # noqa
                        gl.glVertex3f(cell_x+grid_size, cell_y+        0, 0)  # noqa
                        gl.glVertex3f(cell_x+        0, cell_y+        0, 0)  # noqa
                        gl.glEnd()
                w_offset += ring_size + padding
            if save_to_file:
                pyglet.image.get_buffer_manager().get_color_buffer().save(
                    "/tmp/encodeder_rings{:04d}.png".format(self.rendering_iteration))
            # actualize
            win.flip()
            self.rendering_iteration += 1
            return self.viewer.isopen

class EncodedEnv(IANEnv):
    """ takes a (2) action as input
    outputs encoded obs (546) """
    def __init__(self, backend, encoding,
                 silent=False, max_episode_length=1000, collect_trajectories=False,
                 gpu=False, encoder=None):
        if encoder is None:
            encoder = EnvEncoder(backend, encoding)
        self.encoder = encoder
        super(EncodedEnv, self).__init__(
            silent=silent, max_episode_length=max_episode_length, collect_trajectories=collect_trajectories)
        self.action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        self.observation_space = self.encoder.observation_space

    def step(self, action):
        action = np.array([action[0], action[1], 0.])  # no rotation
        obs, reward, done, info = super(EncodedEnv, self).step(action)
        h = self.encoder._encode_obs(obs, action)
        return h, reward, done, info

    def reset(self):
        self.encoder.reset()
        obs = super(EncodedEnv, self).reset()
        h = self.encoder._encode_obs(obs, np.array([0,0,0]))
        return h

    def render(self, mode="human", close=False, save_to_file=False, robocentric=False,
               render_decoded_scan=True):
        if mode == "rings":
            self.encoder._render_rings(close=close)
            return
        decoded_scan = None
        if render_decoded_scan:
            decoded_scan = self.encoder._get_last_decoded_scan()
        super(EncodedEnv, self).render(mode=mode, close=close, lidar_scan_override=decoded_scan,
                                       save_to_file=save_to_file, robocentric=robocentric)



from __future__ import print_function
import numpy as np
import os
import json
import tensorflow as tf
import random
import time
from timeit import default_timer as timer

from navrep.tools.rings import generate_rings
from navrep.models.rnn import reset_graph, sample_hps_params, MDNRNN, get_pi_idx
from navrep.models.vae2d import ConvVAE

# parameters
TEMPERATURE = 0.5
_Z = 32

sequence_z_path = os.path.expanduser(
    "~/navrep/datasets/M/ian/000_mus_logvars_robotstates_actions_rewards_dones.npz"
)
rnn_model_path = os.path.expanduser("~/navrep/models/M/rnn.json")
vae_model_path = os.path.expanduser("~/navrep/models/V/vae.json")

reset_graph()
rnn = MDNRNN(sample_hps_params, gpu_mode=False)
vae = ConvVAE(batch_size=1, is_training=False)

vae.load_json(vae_model_path)
rnn.load_json(rnn_model_path)

rings_def = generate_rings(64, 64)

# load sequence image encoding
arrays = np.load(sequence_z_path)
sequence_action = arrays["actions"]
sequence_mu = arrays["mus"]
sequence_logvar = arrays["logvars"]
sequence_z = sequence_mu + np.exp(sequence_logvar / 2.0) * np.random.randn(
    *(sequence_mu.shape)
)
SEQUENCE_LENGTH = len(sequence_mu)

prev_z = sequence_z[0]
prev_z_predicted = sequence_z[0]
prev_action = sequence_action[0]
prev_restart = np.array([0])
rnn_state = rnn.sess.run(rnn.zero_state)
for i in range(SEQUENCE_LENGTH):
    print(prev_action)

    feed = {
        rnn.input_z: np.reshape(prev_z, (1, 1, _Z)),
        rnn.input_action: np.reshape(prev_action, (1, 1, 3)),
        rnn.input_restart: np.reshape(prev_restart, (1, 1)),
        rnn.initial_state: rnn_state,
    }

    [logmix, mean, logstd, logrestart, next_state] = rnn.sess.run(
        [
            rnn.out_logmix,
            rnn.out_mean,
            rnn.out_logstd,
            rnn.out_restart_logits,
            rnn.final_state,
        ],
        feed,
    )

    OUTWIDTH = _Z

    if TEMPERATURE == 0:  # deterministically pick max of MDN distribution
        mixture_idx = np.argmax(logmix, axis=-1)
        chosen_mean = mean[(range(OUTWIDTH), mixture_idx)]
        chosen_logstd = logstd[(range(OUTWIDTH), mixture_idx)]
        next_z_predicted = chosen_mean
    else:  # sample from modelled MDN distribution
        logmix2 = np.copy(logmix) / TEMPERATURE  # adjust temperatures
        logmix2 -= logmix2.max()
        logmix2 = np.exp(logmix2)
        logmix2 /= logmix2.sum(axis=1).reshape(OUTWIDTH, 1)

        mixture_idx = np.zeros(OUTWIDTH)
        chosen_mean = np.zeros(OUTWIDTH)
        chosen_logstd = np.zeros(OUTWIDTH)
        for j in range(OUTWIDTH):
            idx = get_pi_idx(np.random.rand(), logmix2[j])
            mixture_idx[j] = idx
            chosen_mean[j] = mean[j][idx]
            chosen_logstd[j] = logstd[j][idx]

        rand_gaussian = np.random.randn(OUTWIDTH) * np.sqrt(TEMPERATURE)
        next_z_predicted = chosen_mean + np.exp(chosen_logstd) * rand_gaussian
    if sample_hps_params.differential_z:
        next_z_predicted = prev_z + next_z_predicted  # if in the residual regime

    # take z from sequence
    next_z = sequence_z[i]
    prev_action = sequence_action[i]

    next_restart = 0
    done = False
    if logrestart[0] > 0:
        next_restart = 1
        done = True

    rings_pred = vae.decode(prev_z.reshape(1, _Z)) * rings_def["rings_to_bool"]
    rings_z_pred_pred = (
        vae.decode(prev_z_predicted.reshape(1, _Z)) * rings_def["rings_to_bool"]
    )
    predicted_ranges = rings_def["rings_to_lidar"](rings_pred, 1080)
    if True:
        import matplotlib.pyplot as plt

        plt.ion()
        plt.figure("rings")
        plt.clf()
        fig, (ax1, ax2) = plt.subplots(
            2, 1, subplot_kw=dict(projection="polar"), num="rings"
        )
        rings_def["visualize_rings"](rings_pred[0], scan=None, fig=fig, ax=ax1)
        rings_def["visualize_rings"](rings_z_pred_pred[0], scan=None, fig=fig, ax=ax2)
        ax1.set_ylim([0, 10])
        ax1.set_title(i)
        ax2.set_ylim([0, 10])
        # update
        plt.pause(0.01)

    prev_z = next_z
    prev_z_predicted = next_z_predicted
    prev_restart = next_restart
    rnn_state = next_state



from __future__ import print_function
import os
import threading
import numpy as np
import gym
from gym import spaces

from navrep.tools.wdataset import WorldModelDataset
from navrep.envs.encodedenv import EnvEncoder

MAX_LIDAR_DIST = 25.  # used here for rendering purposes only

class ArchiveEnv(gym.Env):
    """ This class wraps the IARLenv to make a RL ready simplified environment
    """
    metadata = {'render.modes': ['human']}

    def __init__(self, directory, file_limit=None, silent=False, max_episode_length=1000):
        # gym env definition
        super(ArchiveEnv, self).__init__()
        self.action_space = spaces.Box(low=-1, high=1, shape=(3,), dtype=np.float32)
        self.observation_space = spaces.Tuple((
            spaces.Box(low=-np.inf, high=np.inf, shape=(1080,), dtype=np.float32),
            spaces.Box(low=-np.inf, high=np.inf, shape=(5,), dtype=np.float32),
        ))
        self.data = WorldModelDataset._load_data(None, directory, file_limit=file_limit)
        if len(self.data["scans"]) == 0:
            raise ValueError
        self.current_iteration = None
        self.data["dones"][-1] = 1
        self.viewer = None

    def _step(self, action):
        """ preserved when inherited classes overrite step() """
        scan = self.data["scans"][self.current_iteration].astype(np.float32)
        robotstate = self.data["robotstates"][self.current_iteration]
        reward = self.data["rewards"][self.current_iteration]
        done = self.data["dones"][self.current_iteration]
        obs = (scan, robotstate)  # latest scan only (buffer, ray, channel)
        if not done:
            self.current_iteration += 1
        return obs, reward, done, {}

    def step(self, action):
        return self._step(action)

    def reset(self):
        action = self.data["actions"][self.current_iteration]
        if self.current_iteration is None:
            self.current_iteration = 0
            obs, _, _, _ = self._step(action)
            return obs
        # skip to the start of next episode
        while True:
            obs, _, done, _ = self._step(action)
            if done:
                self.current_iteration += 1
                if self.current_iteration >= len(self.data["scans"]):
                    self.current_iteration = 0
                break
        return obs

    def close(self):
        self.render(close=True)

    def _get_viewer(self):
        return self.viewer

    def _get_dt(self):
        return 0.25

    def render(self, mode="human", close=False, lidar_scan_override=None, save_to_file=False,
               draw_score=True):
        if close:
            if self.viewer is not None:
                self.viewer.close()
            return

        # get last obs
        goal_pred = self.data["robotstates"][self.current_iteration-1][:2]
        action = self.data["actions"][self.current_iteration-1]
        scan = self.data["scans"][self.current_iteration-1]
        scan[scan == 0] = MAX_LIDAR_DIST  # render 0 returns as inf
        if lidar_scan_override is not None:
            scan = lidar_scan_override

        if mode == "rgb_array":
            raise NotImplementedError
        elif mode in ["human", "rings"]:
            # Window and viewport size
            WINDOW_W = 256
            WINDOW_H = 256
            M_PER_PX = 13.6 / WINDOW_H
            VP_W = WINDOW_W
            VP_H = WINDOW_H
            from gym.envs.classic_control import rendering
            import pyglet
            from pyglet import gl

            # enable this to render for hg_archivenv visual
            if False:
                draw_score = False
                save_to_file = True
                WINDOW_W = 512
                WINDOW_H = 512
                M_PER_PX = 51.2 / WINDOW_H

            # Create viewer
            if self.viewer is None:
                self.viewer = rendering.Viewer(WINDOW_W, WINDOW_H)
                self.score_label = pyglet.text.Label(
                    "0000",
                    font_size=12,
                    x=20,
                    y=WINDOW_H * 2.5 / 40.00,
                    anchor_x="left",
                    anchor_y="center",
                    color=(255, 255, 255, 255),
                )
                #                 self.transform = rendering.Transform()
                self.currently_rendering_iteration = 0
                self.image_lock = threading.Lock()
            # Render in pyglet
            def make_circle(c, r, res=10):
                thetas = np.linspace(0, 2 * np.pi, res + 1)[:-1]
                verts = np.zeros((res, 2))
                verts[:, 0] = c[0] + r * np.cos(thetas)
                verts[:, 1] = c[1] + r * np.sin(thetas)
                return verts

            with self.image_lock:
                self.currently_rendering_iteration += 1
                self.viewer.draw_circle(r=10, color=(0.3, 0.3, 0.3))
                win = self.viewer.window
                win.switch_to()
                win.dispatch_events()
                win.clear()
                gl.glViewport(0, 0, VP_W, VP_H)
                # colors
#                 bgcolor = np.array([0.4, 0.8, 0.4])
                bgcolor = np.array([0.7, 0.75, 0.86])
                nosecolor = np.array([0.3, 0.3, 0.3])
#                 lidarcolor = np.array([1.0, 0.0, 0.0])
                lidarcolor = np.array([1., 0.3, 0.0])
#                 fovcolor = np.array([1., 1., 0.])
                fovcolor = np.array([0.8, 0.8, 0.2])
                # Green background
                gl.glBegin(gl.GL_QUADS)
                gl.glColor4f(bgcolor[0], bgcolor[1], bgcolor[2], 1.0)
                gl.glVertex3f(0, VP_H, 0)
                gl.glVertex3f(VP_W, VP_H, 0)
                gl.glVertex3f(VP_W, 0, 0)
                gl.glVertex3f(0, 0, 0)
                gl.glEnd()
                # LIDAR
                i = WINDOW_W / 2.0
                j = WINDOW_H / 2.0
                angle = np.pi / 2.0
                lidar_angles = np.linspace(0, 2 * np.pi, len(scan) + 1)[:-1]
                lidar_angles = lidar_angles + np.pi / 2.  # make robot face up
                i_ray_ends = i + scan / M_PER_PX * np.cos(lidar_angles)
                j_ray_ends = j + scan / M_PER_PX * np.sin(lidar_angles)
                is_in_fov = np.cos(lidar_angles - angle) >= 0.78
                for ray_idx in range(len(scan)):
                    end_i = i_ray_ends[ray_idx]
                    end_j = j_ray_ends[ray_idx]
                    gl.glBegin(gl.GL_LINE_LOOP)
                    if is_in_fov[ray_idx]:
                        gl.glColor4f(fovcolor[0], fovcolor[1], fovcolor[2], 0.1)
                    else:
                        gl.glColor4f(lidarcolor[0], lidarcolor[1], lidarcolor[2], 0.1)
                    gl.glVertex3f(i, j, 0)
                    gl.glVertex3f(end_i, end_j, 0)
                    gl.glEnd()
                # Agent body
                i = WINDOW_W / 2.0
                j = WINDOW_H / 2.0
                r = 0.3 / M_PER_PX
                angle = np.pi / 2.0
                poly = make_circle((i, j), r)
                gl.glBegin(gl.GL_POLYGON)
                color = np.array([1.0, 1.0, 1.0])
                gl.glColor4f(color[0], color[1], color[2], 1)
                for vert in poly:
                    gl.glVertex3f(vert[0], vert[1], 0)
                gl.glEnd()
                # Direction triangle
                inose = i + r * np.cos(angle)
                jnose = j + r * np.sin(angle)
                iright = i + 0.3 * r * -np.sin(angle)
                jright = j + 0.3 * r * np.cos(angle)
                ileft = i - 0.3 * r * -np.sin(angle)
                jleft = j - 0.3 * r * np.cos(angle)
                gl.glBegin(gl.GL_TRIANGLES)
                gl.glColor4f(nosecolor[0], nosecolor[1], nosecolor[2], 1)
                gl.glVertex3f(inose, jnose, 0)
                gl.glVertex3f(iright, jright, 0)
                gl.glVertex3f(ileft, jleft, 0)
                gl.glEnd()
                # Goal
                goalcolor = np.array([1., 1., 0.3])
                px_goal = goal_pred / M_PER_PX
                igoal = i - px_goal[1]  # rotate 90deg to face up
                jgoal = j + px_goal[0]
                # Goal line
                gl.glBegin(gl.GL_LINE_LOOP)
                gl.glColor4f(goalcolor[0], goalcolor[1], goalcolor[2], 1)
                gl.glVertex3f(i, j, 0)
                gl.glVertex3f(igoal, jgoal, 0)
                gl.glEnd()
                # Goal markers
                gl.glBegin(gl.GL_TRIANGLES)
                gl.glColor4f(goalcolor[0], goalcolor[1], goalcolor[2], 1)
                triangle = make_circle((igoal, jgoal), r, res=3)
                for vert in triangle:
                    gl.glVertex3f(vert[0], vert[1], 0)
                gl.glEnd()
                # Text
                self.score_label.text = "File {} It {} a {:.1f} {:.1f} {:.1f}".format(
                    self.current_iteration // 1000,
                    self.current_iteration % 1000,
                    action[0],
                    action[1],
                    action[2],
                )
                if draw_score:
                    self.score_label.draw()
                win.flip()
                if save_to_file:
                    pyglet.image.get_buffer_manager().get_color_buffer().save(
                        "/tmp/archivenv{:05}.png".format(self.currently_rendering_iteration))
                return self.viewer.isopen


class ArchiveEncoder(EnvEncoder):
    def __init__(self, backend, encoding):
        assert backend == "VAE_LSTM"
        super(ArchiveEncoder, self).__init__(
            backend, encoding,
            rnn_model_path=os.path.expanduser("~/navrep/models/M/navreptrainrnn.json"),
            vae_model_path=os.path.expanduser("~/navrep/models/V/navreptrainvae.json"),
        )


class ArchiveEncodedEnv(ArchiveEnv):
    """ takes a (2) action as input
    outputs encoded obs (546) """
    def __init__(self, backend, encoding,
                 directory, file_limit=None, silent=False, max_episode_length=1000):
        self.encoder = ArchiveEncoder(backend, encoding)
        super(ArchiveEncodedEnv, self).__init__(directory, file_limit=file_limit, silent=silent,
                                                max_episode_length=max_episode_length)
        self.action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        self.observation_space = self.encoder.observation_space

    def step(self, action):
        action = np.array([action[0], action[1], 0.])  # no rotation
        obs, reward, done, info = super(ArchiveEncodedEnv, self).step(action)
        h = self.encoder._encode_obs(obs, action)
        return h, reward, done, info

    def reset(self, *args, **kwargs):
        self.encoder.reset()
        obs = super(ArchiveEncodedEnv, self).reset(*args, **kwargs)
        h = self.encoder._encode_obs(obs, np.array([0,0,0]))
        return h

    def close(self):
        super(ArchiveEncodedEnv, self).close()
        self.encoder.close()

    def render(self, mode="human", close=False, save_to_file=False):
        decoded_scan = self.encoder._get_last_decoded_scan()
        super(ArchiveEncodedEnv, self).render(
            mode=mode, close=close, lidar_scan_override=decoded_scan, save_to_file=save_to_file)
        if mode == "rings":
            self.encoder._render_rings(close=close, save_to_file=save_to_file)



from __future__ import print_function
import numpy as np
import os
import threading

from navrep.models.rnn import reset_graph, sample_hps_params, MDNRNN, get_pi_idx
from navrep.models.vae2d import ConvVAE

_64 = 64  # size of VAE image input
_Z = 32  # size of VAE z latent space

class ImDreamEnv(object):
    def __init__(self, temperature=0.25):
        # constants
        self.TEMPERATURE = temperature
        self.DT = 0.2  # should be the same as data rnn was trained with
        initial_z_path = os.path.expanduser(
            "~/navrep/datasets/M/im/corridor_koze_kids_bag_mus_logvars_robotstates_actions_rewards_dones.npz"
        )
        rnn_model_path = os.path.expanduser("~/navrep/models/M/imrnn.json")
        vae_model_path = os.path.expanduser("~/navrep/models/V/imvae.json")
        # V + M Models
        reset_graph()
        self.rnn = MDNRNN(sample_hps_params, gpu_mode=False)
        self.vae = ConvVAE(batch_size=1, is_training=False, channels=3)
        self.vae.load_json(vae_model_path)
        self.rnn.load_json(rnn_model_path)
        # load initial image encoding
        arrays = np.load(initial_z_path)
        initial_mu = arrays["mus"][0]
        initial_logvar = arrays["logvars"][0]
        self.initial_z = initial_mu + np.exp(initial_logvar / 2.0) * np.random.randn(
            *(initial_mu.shape)
        )
        # other tools
        self.viewer = None
        # environment state variables
        self.reset()
        # hot-start the rnn state
        for i in range(20):
            self.step(np.array([0,0,0]), override_next_z=self.initial_z)

    def step(self, action, override_next_z=None):
        feed = {
            self.rnn.input_z: np.reshape(self.prev_z, (1, 1, _Z)),
            self.rnn.input_action: np.reshape(action, (1, 1, 3)),
            self.rnn.input_restart: np.reshape(self.prev_restart, (1, 1)),
            self.rnn.initial_state: self.rnn_state,
        }

        [logmix, mean, logstd, logrestart, next_state] = self.rnn.sess.run(
            [
                self.rnn.out_logmix,
                self.rnn.out_mean,
                self.rnn.out_logstd,
                self.rnn.out_restart_logits,
                self.rnn.final_state,
            ],
            feed,
        )
        OUTWIDTH = _Z

        if self.TEMPERATURE == 0:  # deterministically pick max of MDN distribution
            mixture_idx = np.argmax(logmix, axis=-1)
            chosen_mean = mean[(range(OUTWIDTH), mixture_idx)]
            chosen_logstd = logstd[(range(OUTWIDTH), mixture_idx)]
            next_z = chosen_mean
        else:  # sample from modelled MDN distribution
            mixprob = np.copy(logmix) / self.TEMPERATURE  # adjust temperatures
            mixprob -= mixprob.max()
            mixprob = np.exp(mixprob)
            mixprob /= mixprob.sum(axis=1).reshape(OUTWIDTH, 1)

            mixture_idx = np.zeros(OUTWIDTH)
            chosen_mean = np.zeros(OUTWIDTH)
            chosen_logstd = np.zeros(OUTWIDTH)
            for j in range(OUTWIDTH):
                idx = get_pi_idx(np.random.rand(), mixprob[j])
                mixture_idx[j] = idx
                chosen_mean[j] = mean[j][idx]
                chosen_logstd[j] = logstd[j][idx]
            rand_gaussian = np.random.randn(OUTWIDTH) * np.sqrt(self.TEMPERATURE)
            next_z = chosen_mean + np.exp(chosen_logstd) * rand_gaussian
        if sample_hps_params.differential_z:
            next_z = self.prev_z + next_z

        next_restart = 0
        #         if logrestart[0] > 0:
        #             next_restart = 1

        self.prev_z = next_z
        if override_next_z is not None:
            self.prev_z = override_next_z
        self.prev_restart = next_restart
        self.rnn_state = next_state
        # logging-only vars, used for rendering
        self.prev_action = action
        self.episode_step += 1

        return next_z, None, next_restart, {}

    def reset(self):
        self.prev_z = self.initial_z
        self.prev_restart = np.array([1])
        self.rnn_state = self.rnn.sess.run(self.rnn.zero_state)
        # logging vars
        self.prev_action = np.array([0.0, 0.0, 0.0])
        self.episode_step = 0

    def render(self, mode="human", close=False):
        img_pred = (self.vae.decode(self.prev_z.reshape(1, _Z)) * 255).astype(np.uint8)
        img_pred = img_pred.reshape(_64, _64, 3)

        if mode == "rgb_array":
            raise NotImplementedError
        elif mode == "human":
            # Window and viewport size
            WINDOW_W = 256
            WINDOW_H = 256
            VP_W = WINDOW_W
            VP_H = WINDOW_H
            from gym.envs.classic_control import rendering
            import pyglet
            from pyglet import gl

            # Create pyglet image
#             pixels = [
#                 255, 0, 0,      0, 255, 0,      0, 0, 255,     # RGB values range from
#                 255, 0, 0,      255, 0, 0,      255, 0, 0,     # 0 to 255 for each color
#                 255, 0, 0,      255, 0, 0,      255, 0, 0,     # component.
#             ]
            from pyglet.gl.gl import GLubyte
            pixels = img_pred.flatten()
            rawData = (GLubyte * len(pixels))(*pixels)
            image_data = pyglet.image.ImageData(_64, _64, 'RGB', rawData)

            # Create viewer
            if self.viewer is None:
                self.viewer = rendering.Viewer(WINDOW_W, WINDOW_H)
                self.score_label = pyglet.text.Label(
                    "0000",
                    font_size=12,
                    x=20,
                    y=WINDOW_H * 2.5 / 40.00,
                    anchor_x="left",
                    anchor_y="center",
                    color=(255, 255, 255, 255),
                )
                #                 self.transform = rendering.Transform()
                self.currently_rendering_iteration = 0
                self.image_lock = threading.Lock()
            # Render in pyglet
            with self.image_lock:
                self.currently_rendering_iteration += 1
                self.viewer.draw_circle(r=10, color=(0.3, 0.3, 0.3))
                win = self.viewer.window
                win.switch_to()
                win.dispatch_events()
                win.clear()
                gl.glViewport(0, 0, VP_W, VP_H)
                # Image
                image_data.blit(96,96)
                # Text
                self.score_label.text = "A {:.1f} {:.1f} {:.1f} S {}".format(
                    self.prev_action[0],
                    self.prev_action[1],
                    self.prev_action[2],
                    self.episode_step,
                )
                self.score_label.draw()
                win.flip()
                return self.viewer.isopen




# This Python file uses the following encoding: utf-8
"""autogenerated by genpy from frame_msgs/TrackedPerson2d.msg. Do not edit."""
import sys
python3 = True if sys.hexversion > 0x03000000 else False
import genpy
import struct


class TrackedPerson2d(genpy.Message):
  _md5sum = "12df4823c658df0d660d2a5c68ae4aea"
  _type = "frame_msgs/TrackedPerson2d"
  _has_header = False #flag to mark the presence of a Header object
  _full_text = """# Message defining a 2d image bbox of a tracked person
#

uint64      track_id        # unique identifier of the target, consistent over time
float32     person_height   # 3d height of person in m
int32       x               # top left corner x of 2d image bbox
int32       y               # top left corner y of 2d image bbox
uint32      w               # width of 2d image bbox
uint32      h               # height of 2d image bbox
float32     depth           # distance from the camera in m
"""
  __slots__ = ['track_id','person_height','x','y','w','h','depth']
  _slot_types = ['uint64','float32','int32','int32','uint32','uint32','float32']

  def __init__(self, *args, **kwds):
    """
    Constructor. Any message fields that are implicitly/explicitly
    set to None will be assigned a default value. The recommend
    use is keyword arguments as this is more robust to future message
    changes.  You cannot mix in-order arguments and keyword arguments.

    The available fields are:
       track_id,person_height,x,y,w,h,depth

    :param args: complete set of field values, in .msg order
    :param kwds: use keyword arguments corresponding to message field names
    to set specific fields.
    """
    if args or kwds:
      super(TrackedPerson2d, self).__init__(*args, **kwds)
      #message fields cannot be None, assign default values for those that are
      if self.track_id is None:
        self.track_id = 0
      if self.person_height is None:
        self.person_height = 0.
      if self.x is None:
        self.x = 0
      if self.y is None:
        self.y = 0
      if self.w is None:
        self.w = 0
      if self.h is None:
        self.h = 0
      if self.depth is None:
        self.depth = 0.
    else:
      self.track_id = 0
      self.person_height = 0.
      self.x = 0
      self.y = 0
      self.w = 0
      self.h = 0
      self.depth = 0.

  def _get_types(self):
    """
    internal API method
    """
    return self._slot_types

  def serialize(self, buff):
    """
    serialize message into buffer
    :param buff: buffer, ``StringIO``
    """
    try:
      _x = self
      buff.write(_get_struct_Qf2i2If().pack(_x.track_id, _x.person_height, _x.x, _x.y, _x.w, _x.h, _x.depth))
    except struct.error as se: self._check_types(struct.error("%s: '%s' when writing '%s'" % (type(se), str(se), str(locals().get('_x', self)))))
    except TypeError as te: self._check_types(ValueError("%s: '%s' when writing '%s'" % (type(te), str(te), str(locals().get('_x', self)))))

  def deserialize(self, str):
    """
    unpack serialized message in str into this message instance
    :param str: byte array of serialized message, ``str``
    """
    try:
      end = 0
      _x = self
      start = end
      end += 32
      (_x.track_id, _x.person_height, _x.x, _x.y, _x.w, _x.h, _x.depth,) = _get_struct_Qf2i2If().unpack(str[start:end])
      return self
    except struct.error as e:
      raise genpy.DeserializationError(e) #most likely buffer underfill


  def serialize_numpy(self, buff, numpy):
    """
    serialize message with numpy array types into buffer
    :param buff: buffer, ``StringIO``
    :param numpy: numpy python module
    """
    try:
      _x = self
      buff.write(_get_struct_Qf2i2If().pack(_x.track_id, _x.person_height, _x.x, _x.y, _x.w, _x.h, _x.depth))
    except struct.error as se: self._check_types(struct.error("%s: '%s' when writing '%s'" % (type(se), str(se), str(locals().get('_x', self)))))
    except TypeError as te: self._check_types(ValueError("%s: '%s' when writing '%s'" % (type(te), str(te), str(locals().get('_x', self)))))

  def deserialize_numpy(self, str, numpy):
    """
    unpack serialized message in str into this message instance using numpy for array types
    :param str: byte array of serialized message, ``str``
    :param numpy: numpy python module
    """
    try:
      end = 0
      _x = self
      start = end
      end += 32
      (_x.track_id, _x.person_height, _x.x, _x.y, _x.w, _x.h, _x.depth,) = _get_struct_Qf2i2If().unpack(str[start:end])
      return self
    except struct.error as e:
      raise genpy.DeserializationError(e) #most likely buffer underfill

_struct_I = genpy.struct_I
def _get_struct_I():
    global _struct_I
    return _struct_I
_struct_Qf2i2If = None
def _get_struct_Qf2i2If():
    global _struct_Qf2i2If
    if _struct_Qf2i2If is None:
        _struct_Qf2i2If = struct.Struct("<Qf2i2If")
    return _struct_Qf2i2If



from __future__ import print_function
import numpy as np
import os
import threading

from navrep.tools.rings import generate_rings
from navrep.models.rnn import reset_graph, sample_hps_params, MDNRNN, get_pi_idx, MAX_GOAL_DIST
from navrep.models.vae2d import ConvVAE

_Z = 32
_G = 2

class DreamEnv(object):
    def __init__(self, temperature=0.25,
                 initial_z_path=os.path.expanduser(
                     "~/navrep/datasets/M/ian/000_mus_logvars_robotstates_actions_rewards_dones.npz"
                 ),
                 rnn_model_path=os.path.expanduser("~/navrep/models/M/rnn.json"),
                 vae_model_path=os.path.expanduser("~/navrep/models/V/vae.json"),
                 ):
        # constants
        self.TEMPERATURE = temperature
        self.DT = 0.5  # should be the same as data rnn was trained with
        # V + M Models
        reset_graph()
        self.rnn = MDNRNN(sample_hps_params, gpu_mode=False)
        self.vae = ConvVAE(batch_size=1, is_training=False)
        self.vae.load_json(vae_model_path)
        self.rnn.load_json(rnn_model_path)
        # load initial image encoding
        arrays = np.load(initial_z_path)
        initial_mu = arrays["mus"][0]
        initial_logvar = arrays["logvars"][0]
        initial_robotstate = arrays["robotstates"][0]
        ini_lidar_z = initial_mu + np.exp(initial_logvar / 2.0) * np.random.randn(
            *(initial_mu.shape)
        )
        ini_goal_z = initial_robotstate[:2] / MAX_GOAL_DIST
        self.initial_z = np.concatenate([ini_lidar_z, ini_goal_z], axis=-1)
        # other tools
        self.rings_def = generate_rings(64, 64)
        self.viewer = None
        # environment state variables
        self.reset()
        # hot-start the rnn state
        for i in range(20):
            self.step(np.array([0,0,0]), override_next_z=self.initial_z)

    def step(self, action, override_next_z=None):
        feed = {
            self.rnn.input_z: np.reshape(self.prev_z, (1, 1, _Z+_G)),
            self.rnn.input_action: np.reshape(action, (1, 1, 3)),
            self.rnn.input_restart: np.reshape(self.prev_restart, (1, 1)),
            self.rnn.initial_state: self.rnn_state,
        }

        [logmix, mean, logstd, logrestart, next_state] = self.rnn.sess.run(
            [
                self.rnn.out_logmix,
                self.rnn.out_mean,
                self.rnn.out_logstd,
                self.rnn.out_restart_logits,
                self.rnn.final_state,
            ],
            feed,
        )
        OUTWIDTH = _Z+_G

        if self.TEMPERATURE == 0:  # deterministically pick max of MDN distribution
            mixture_idx = np.argmax(logmix, axis=-1)
            chosen_mean = mean[(range(OUTWIDTH), mixture_idx)]
            chosen_logstd = logstd[(range(OUTWIDTH), mixture_idx)]
            next_z = chosen_mean
        else:  # sample from modelled MDN distribution
            mixprob = np.copy(logmix) / self.TEMPERATURE  # adjust temperatures
            mixprob -= mixprob.max()
            mixprob = np.exp(mixprob)
            mixprob /= mixprob.sum(axis=1).reshape(OUTWIDTH, 1)

            mixture_idx = np.zeros(OUTWIDTH)
            chosen_mean = np.zeros(OUTWIDTH)
            chosen_logstd = np.zeros(OUTWIDTH)
            for j in range(OUTWIDTH):
                idx = get_pi_idx(np.random.rand(), mixprob[j])
                mixture_idx[j] = idx
                chosen_mean[j] = mean[j][idx]
                chosen_logstd[j] = logstd[j][idx]
            rand_gaussian = np.random.randn(OUTWIDTH) * np.sqrt(self.TEMPERATURE)
            next_z = chosen_mean + np.exp(chosen_logstd) * rand_gaussian
        if sample_hps_params.differential_z:
            next_z = self.prev_z + next_z

        next_restart = 0
        #         if logrestart[0] > 0:
        #             next_restart = 1

        self.prev_z = next_z
        if override_next_z is not None:
            self.prev_z = override_next_z
        self.prev_restart = next_restart
        self.rnn_state = next_state
        # logging-only vars, used for rendering
        self.prev_action = action
        self.episode_step += 1

        return next_z, None, next_restart, {}

    def reset(self):
        self.prev_z = self.initial_z
        self.prev_restart = np.array([1])
        self.rnn_state = self.rnn.sess.run(self.rnn.zero_state)
        # logging vars
        self.prev_action = np.array([0.0, 0.0, 0.0])
        self.episode_step = 0

    def render(self, mode="human", close=False):
        if close:
            if self.viewer is not None:
                self.viewer.close()
            return

        # get last z decoding
        rings_pred = (
            self.vae.decode(self.prev_z.reshape(1, _Z+_G)[:, :_Z])
            * self.rings_def["rings_to_bool"]
        )
        predicted_ranges = self.rings_def["rings_to_lidar"](rings_pred, 1080)
        goal_pred = self.prev_z.reshape((_Z+_G,))[_Z:] * MAX_GOAL_DIST

        if mode == "rgb_array":
            raise NotImplementedError
        elif mode == "human":
            # Window and viewport size
            WINDOW_W = 256
            WINDOW_H = 256
            M_PER_PX = 25.6 / WINDOW_H
            VP_W = WINDOW_W
            VP_H = WINDOW_H
            from gym.envs.classic_control import rendering
            import pyglet
            from pyglet import gl

            # Create viewer
            if self.viewer is None:
                self.viewer = rendering.Viewer(WINDOW_W, WINDOW_H)
                self.score_label = pyglet.text.Label(
                    "0000",
                    font_size=12,
                    x=20,
                    y=WINDOW_H * 2.5 / 40.00,
                    anchor_x="left",
                    anchor_y="center",
                    color=(255, 255, 255, 255),
                )
                #                 self.transform = rendering.Transform()
                self.currently_rendering_iteration = 0
                self.image_lock = threading.Lock()
            # Render in pyglet
            def make_circle(c, r, res=10):
                thetas = np.linspace(0, 2 * np.pi, res + 1)[:-1]
                verts = np.zeros((res, 2))
                verts[:, 0] = c[0] + r * np.cos(thetas)
                verts[:, 1] = c[1] + r * np.sin(thetas)
                return verts

            with self.image_lock:
                self.currently_rendering_iteration += 1
                self.viewer.draw_circle(r=10, color=(0.3, 0.3, 0.3))
                win = self.viewer.window
                win.switch_to()
                win.dispatch_events()
                win.clear()
                gl.glViewport(0, 0, VP_W, VP_H)
                # colors
                bgcolor = np.array([0.4, 0.8, 0.4])
                nosecolor = np.array([0.3, 0.3, 0.3])
                lidarcolor = np.array([1.0, 0.0, 0.0])
                # Green background
                gl.glBegin(gl.GL_QUADS)
                gl.glColor4f(bgcolor[0], bgcolor[1], bgcolor[2], 1.0)
                gl.glVertex3f(0, VP_H, 0)
                gl.glVertex3f(VP_W, VP_H, 0)
                gl.glVertex3f(VP_W, 0, 0)
                gl.glVertex3f(0, 0, 0)
                gl.glEnd()
                # LIDAR
                i = WINDOW_W / 2.0
                j = WINDOW_H / 2.0
                angle = np.pi / 2.0
                scan = np.squeeze(predicted_ranges)
                lidar_angles = np.linspace(0, 2 * np.pi, len(scan) + 1)[:-1]
                lidar_angles = lidar_angles + np.pi / 2.  # make robot face up
                i_ray_ends = i + scan / M_PER_PX * np.cos(lidar_angles)
                j_ray_ends = j + scan / M_PER_PX * np.sin(lidar_angles)
                is_in_fov = np.cos(lidar_angles - angle) >= 0.78
                for ray_idx in range(len(scan)):
                    end_i = i_ray_ends[ray_idx]
                    end_j = j_ray_ends[ray_idx]
                    gl.glBegin(gl.GL_LINE_LOOP)
                    if is_in_fov[ray_idx]:
                        gl.glColor4f(1.0, 1.0, 0.0, 0.1)
                    else:
                        gl.glColor4f(lidarcolor[0], lidarcolor[1], lidarcolor[2], 0.1)
                    gl.glVertex3f(i, j, 0)
                    gl.glVertex3f(end_i, end_j, 0)
                    gl.glEnd()
                # Agent body
                i = WINDOW_W / 2.0
                j = WINDOW_H / 2.0
                r = 0.3 / M_PER_PX
                angle = np.pi / 2.0
                poly = make_circle((i, j), r)
                gl.glBegin(gl.GL_POLYGON)
                color = np.array([1.0, 1.0, 1.0])
                gl.glColor4f(color[0], color[1], color[2], 1)
                for vert in poly:
                    gl.glVertex3f(vert[0], vert[1], 0)
                gl.glEnd()
                # Direction triangle
                inose = i + r * np.cos(angle)
                jnose = j + r * np.sin(angle)
                iright = i + 0.3 * r * -np.sin(angle)
                jright = j + 0.3 * r * np.cos(angle)
                ileft = i - 0.3 * r * -np.sin(angle)
                jleft = j - 0.3 * r * np.cos(angle)
                gl.glBegin(gl.GL_TRIANGLES)
                gl.glColor4f(nosecolor[0], nosecolor[1], nosecolor[2], 1)
                gl.glVertex3f(inose, jnose, 0)
                gl.glVertex3f(iright, jright, 0)
                gl.glVertex3f(ileft, jleft, 0)
                gl.glEnd()
                # Goal
                goalcolor = np.array([1., 1., 0.3])
                px_goal = goal_pred / M_PER_PX
                igoal = i - px_goal[1]  # rotate 90deg to face up
                jgoal = j + px_goal[0]
                # Goal line
                gl.glBegin(gl.GL_LINE_LOOP)
                gl.glColor4f(goalcolor[0], goalcolor[1], goalcolor[2], 1)
                gl.glVertex3f(i, j, 0)
                gl.glVertex3f(igoal, jgoal, 0)
                gl.glEnd()
                # Goal markers
                gl.glBegin(gl.GL_TRIANGLES)
                gl.glColor4f(goalcolor[0], goalcolor[1], goalcolor[2], 1)
                triangle = make_circle((igoal, jgoal), r/3., res=3)
                for vert in triangle:
                    gl.glVertex3f(vert[0], vert[1], 0)
                gl.glEnd()
                # Text
                self.score_label.text = "A {:.1f} {:.1f} {:.1f} S {}".format(
                    self.prev_action[0],
                    self.prev_action[1],
                    self.prev_action[2],
                    self.episode_step,
                )
                self.score_label.draw()
                win.flip()
                return self.viewer.isopen

    def close(self):
        self.render(close=True)

    def _get_dt(self):
        return self.DT

    def _get_viewer(self):
        return self.viewer



import math
import numpy as np
import os
import torch
from torch.utils.data import Dataset

from navrep.tools.rings import generate_rings
from navrep.envs.navreptrainenv import NavRepTrainEnv
from navrep.scripts.make_vae_dataset import generate_vae_dataset, ORCAPolicy

_64 = 64 # rings size
_CH = 1
_G = 2

LIDAR_NORM_FACTOR = 25.  # maximum typical lidar distance, meters

def scans_to_lidar_obs(scans, lidar_mode, rings_def, channel_first):
    if lidar_mode == "rings":
        N, L = scans.shape
        obs = (
            rings_def["lidar_to_rings"](scans).astype(float)
            / rings_def["rings_to_bool"]
        )
        obs[:, :, 0, :] = 0.0 # remove "too close" points
        # swap channel dimension (compat)
        if channel_first:
            obs = np.moveaxis(obs, -1, 1)
    elif lidar_mode == "scans":
        N, L = scans.shape
        obs = scans / LIDAR_NORM_FACTOR
        obs = obs.reshape(N, L, _CH)  # add single channel
        if channel_first:
            obs = np.moveaxis(obs, -1, 1)
    elif lidar_mode == "images":
        obs = scans / 255.
        if channel_first:
            obs = np.moveaxis(obs, -1, 1)
    else:
        raise NotImplementedError
    return obs

class WorldModelDataset(Dataset):
    def __init__(self, directory, sequence_size,
                 file_limit=None,
                 channel_first=True, as_torch_tensors=True,
                 lidar_mode="rings",
                 pre_convert_obs=False,
                 regen=None):
        self.pre_convert_obs = pre_convert_obs
        self.regen = regen
        self.regen_prob = 0.1
        self.lidar_mode = lidar_mode
        self.sequence_size = sequence_size
        self.channel_first = channel_first
        self.as_torch_tensors = as_torch_tensors
        self.rings_def = generate_rings(_64, _64)
        self.data = self._load_data(directory, file_limit=file_limit)
        if self.pre_convert_obs:
            self._preconvert_obs()
        self.regen_head_index = 0  # which part of the data array to regen next
        print("data has %d steps." % (len(self.data["scans"])))

    def _load_data(self, directory, file_limit=None):
        directory = os.path.expanduser(directory)
        # list files
        files = []
        for dirpath, dirnames, filenames in os.walk(directory):
            for filename in [
                f
                for f in filenames
                if f.endswith("scans_robotstates_actions_rewards_dones.npz")
            ]:
                files.append(os.path.join(dirpath, filename))
        if file_limit is None:
            file_limit = len(files)
        data = {
            "scans": [],
            "robotstates": [],
            "actions": [],
            "rewards": [],
            "dones": [],
        }
        arrays_dict = {}
        for path in files[:file_limit]:
            arrays_dict = np.load(path)
            for k in np.load(path):
                data[k].append(arrays_dict[k])
        for k in arrays_dict.keys():
            data[k] = np.concatenate(data[k], axis=0)
        return data

    def _preconvert_obs(self):
        print("Pre-converting scans")
        self.data["obs"] = scans_to_lidar_obs(
            self.data["scans"], self.lidar_mode, self.rings_def, self.channel_first)
        print("Done")

    def __len__(self):
        return math.ceil(len(self.data["scans"]) / (self.sequence_size + 1))

    def __getitem__(self, idx):
        if self.regen is not None and idx == len(self)//2:
            if np.random.random() < self.regen_prob:
                print("PARTIAL REGEN")
                self._partial_regen()
        i = idx * self.sequence_size + 1
        scans = self.data["scans"][i : i + self.sequence_size + 1]
        robotstates = self.data["robotstates"][i : i + self.sequence_size + 1]
        actions = self.data["actions"][i : i + self.sequence_size + 1]
        dones = self.data["dones"][i : i + self.sequence_size + 1]
        if self.pre_convert_obs:
            obs = self.data["obs"][i : i + self.sequence_size + 1]
        # robotstates goal only
        robotstates = robotstates[:, :_G]
        # scans as rings
        if not self.pre_convert_obs:
            obs = scans_to_lidar_obs(scans, self.lidar_mode, self.rings_def, self.channel_first)
        # outputs
        x = obs[:-1] * 1.
        y = obs[1:] * 1.
        actions = actions[:-1]
        x_rs = robotstates[:-1]
        y_rs = robotstates[1:]
        dones = dones[:-1]
        # add "black" frames at episode transition AAABBB -> AA__BB, no need to predict A->B
        x[dones == 1] = 0.0
        y[dones == 1] = 0.0
        # torch
        if self.as_torch_tensors:
            x = torch.tensor(x, dtype=torch.float)
            y = torch.tensor(y, dtype=torch.float)
            actions = torch.tensor(actions, dtype=torch.float)
            x_rs = torch.tensor(x_rs, dtype=torch.float)
            y_rs = torch.tensor(y_rs, dtype=torch.float)
            dones = torch.tensor(dones, dtype=torch.float)
        return x, actions, y, x_rs, y_rs, dones

    def _partial_regen(self, n_new_sequences=1):
        if self.regen == "navreptrain":
            env = NavRepTrainEnv(silent=True, scenario='train', adaptive=False)
            env.soadrl_sim.human_num = 20
            data = generate_vae_dataset(
                env, n_sequences=n_new_sequences,
                policy=ORCAPolicy(suicide_if_stuck=True),
                render=False, archive_dir=None)
            if self.pre_convert_obs:
                data["obs"] = scans_to_lidar_obs(
                    data["scans"], self.lidar_mode, self.rings_def, self.channel_first)
        else:
            print("Regen {} failed".format(self.regen))
            return
        for k in self.data.keys():
            N = len(data[k])  # should be the same for each key
            # check end inside loop to avoid having to pick an arbitrary key
            if self.regen_head_index + N > len(self.data[k]):
                self.regen_head_index = 0
            # replace data
            i = self.regen_head_index
            self.data[k][i : i + N] = data[k]
        self.regen_head_index += N



import numpy as np
from matplotlib import pyplot as plt

from crings import fast_lidar_to_rings


def generate_rings(
    angle_levels=64,
    range_levels=64,
    expansion_term=3.12,
    min_resolution=0.01,
    min_dist=0.3,
    max_dist=25.0,
    VISUALIZE=False,
):
    """
    expansion_term: 0. (equal range levels) 1. (linear increase in depths), >1. (exponential)

    range levels
    ---
    first bin is           0          to   min_dist
    second bin is          min_dist   to   min_resolution
    nth bin is             x          to   x
    ...
    second-to-last bin is  x          to   max_dist
    last bin is            max_dist   to   inf

    bin values
    ---
    0 points in bin     0
    1 points in bin     0.5
    2 points in bin     0.75
    ..
    inf points in bin   1.
    (a.k.a x(n) = 1-2^-n)
    """
    x = np.linspace(0, 1, range_levels - 2)
    expansion_curve = np.power(x, expansion_term)  # a.k.a range_level_depths
    renormalisation_factor = np.sum(expansion_curve) / (
        max_dist - (min_resolution * (range_levels - 2)) - min_dist
    )
    range_level_depths = expansion_curve / renormalisation_factor + min_resolution
    range_level_maxs = np.cumsum(range_level_depths) + min_dist
    range_level_maxs = np.concatenate([[min_dist], range_level_maxs, [np.inf]]).astype(np.float32)
    range_level_mins = np.concatenate([[0.0], range_level_maxs[:-1]]).astype(np.float32)

    if VISUALIZE:
        th = np.linspace(-7.0 / 8.0 * np.pi, 7.0 / 8.0 * np.pi, angle_levels)
        plt.figure("curve")
        plt.plot(range_level_maxs[:-1])
        for x, y in enumerate(range_level_maxs[:-1]):
            plt.axhline(y)
            plt.axvline(x)
        plt.figure("rings")
        for i, r in enumerate(range_level_maxs[:-1]):
            plt.gca().add_artist(plt.Circle((0, 0), r, color="k", fill=False))
        for i in range(angle_levels):
            plt.plot(
                [min_dist * np.cos(th), r * np.cos(th)],
                [min_dist * np.sin(th), r * np.sin(th)],
                "k",
            )
            plt.axis("equal")
        plt.gca().add_artist(plt.Circle((1, 1), 0.3, color="r", zorder=3))
        plt.tight_layout()

    def lidar_to_rings(scans):
        """
        scans: ndarray (n, N_RAYS)   0-100 [m]
        rings: ndarray (n_scans, angle_levels, range_levels, n_channels)
        """
        return fast_lidar_to_rings(scans, angle_levels, range_levels, range_level_mins, range_level_maxs)

    def old_lidar_to_rings(scans):
        """
        scans: ndarray (n, N_RAYS)   0-100 [m]
        rings: ndarray (n_scans, angle_levels, range_levels, n_channels)
        """
        # remove 0 returns
        scans = scans * 1.0
        scans[scans == 0] = np.inf

        CHANNEL = 0
        N_RAYS = scans.shape[1]
        i_to_j, j_to_ii = generate_downsampling_map(N_RAYS, angle_levels)

        rings = np.zeros(
            (scans.shape[0], angle_levels, range_levels, 1), dtype=np.uint8
        )
        # count level hits for each angular section
        # j is index of angular section, ii are indices of corresponding rays
        for j, ii in enumerate(j_to_ii):
            is_unseen = scans[:, ii, None] < range_level_mins[None, None, :]
            is_hit = np.logical_and(
                scans[:, ii, None] < range_level_maxs[None, None, :],
                scans[:, ii, None] >= range_level_mins[None, None, :],
            )
            #             rings[:,j,:,CHANNEL] = np.sum(is_hit, axis=1)
            #             rings[:,j,:,CHANNEL] = 1 * np.all(is_unseen, axis=1) + 2 * np.any(is_hit, axis=1)
            rings[:, j, :, CHANNEL] = np.clip(
                1 * np.any(is_unseen, axis=1) + 2 * np.any(is_hit, axis=1), 0, 2
            )
        return rings

    def rings_to_lidar(rings, N_RAYS=1080):
        CHANNEL = 0
        i_to_j, j_to_ii = generate_downsampling_map(N_RAYS, angle_levels)
        scans = np.zeros((rings.shape[0], N_RAYS), dtype=np.float32)
        for j, ii in enumerate(j_to_ii):
            scans[:, ii] = range_level_mins[np.argmax(rings[:, j, :, :] > 0.4, axis=1)][
                :, None, CHANNEL
            ]
        return scans

    def visualize_rings(
        ring, scan=None, angle_min=0, angle_max=2 * np.pi, fig=None, ax=None, plot_regen=False
    ):
        CHANNEL = 0
        th = np.linspace(angle_min, angle_max, angle_levels)
        r = range_level_mins
        thth, rr = np.meshgrid(th, r)
        if fig is None:
            fig = plt.figure("rings")
        if ax is None:
            ax = fig.add_subplot(111, projection="polar")
        ax.clear()
        ax.pcolormesh(thth, rr, ring[:, :, CHANNEL].T, cmap=plt.cm.Greys)
        if scan is not None:
            scan_regen = rings_to_lidar(ring[None, :, :, :], scan.shape[0])[0, :]
            scan_th = np.linspace(angle_min, angle_max, scan.shape[0])
            plt.plot(scan_th, scan, "r")
            if plot_regen:
                plt.plot(scan_th, scan_regen, "g")
        return ax

    return {
        "range_level_mins": range_level_mins,
        "range_level_maxs": range_level_maxs,
        "lidar_to_rings": lidar_to_rings,
        "rings_to_lidar": rings_to_lidar,
        "visualize_rings": visualize_rings,
        "rings_to_bool": 2.0,
    }


def generate_downsampling_map(I, J):
    """
    with,
    I = 5
    J = 2

    0   1               I
    |___|___|___|___|___|
    |_________|_________|
    0         1         J

    >> i_to_j, j_to_ii = downsample_map(5, 2)
    >> i_to_j
    [0, 0, 0, 1, 1]
    >> j_to_ii[0]
    [0, 1, 2]
    >> j_to_ii[1]
    [3, 4]
    """
    downsample_factor = I * 1.0 / J
    i_to_j = np.floor(np.arange(I) / downsample_factor).astype(int)
    j_to_ii = [
        np.arange(
            np.ceil(j * downsample_factor),
            np.ceil((j + 1) * downsample_factor),
            dtype=int,
        )
        for j in range(J)
    ]
    return i_to_j, j_to_ii


if __name__ == "__main__":
    ring_def = generate_rings()
    # linear ramp
    scans = np.ones((1, 1080)) * (np.arange(1080) / 1080.0 * 25.0)[None, :]
    rings = ring_def["lidar_to_rings"](scans.astype(np.float32))
    plt.ion()
    ring_def["visualize_rings"](rings[0, :, :, :], scan=scans[0])



import gym
import os
from gym import spaces
import numpy as np
from pepper_2d_iarlenv import parse_iaenv_args, IARLEnv, check_iaenv_args
from pkg_resources import resource_filename

class ToyEnv(gym.Env):
    def __init__(self):
        super(ToyEnv, self).__init__()
        self.action_space = spaces.Box(low=-1, high=1, shape=(3,), dtype=np.float32)
        # actually obs space is Tuple(Box(1080), Box(5)) but no support from sb
        self.observation_space = spaces.Tuple((
            spaces.Box(low=0, high=np.inf, shape=(1080,), dtype=np.float32),
            spaces.Box(low=0, high=np.inf, shape=(5,), dtype=np.float32),
        ))
        # args
        args = parse_iaenv_args()
        args.unmerged_scans = False
        args.continuous = True
        args.naive_plan = True
        args.no_ros = True
        args.dt = 0.5
        args.map_folder = resource_filename('asl_pepper_2d_sim_maps', 'maps')
        args.scenario = "toyempty1"
        args.map_name = "empty"
        args.map_downsampling_passes = 0
        check_iaenv_args(args)
        # env
        self.iarlenv = IARLEnv(args)
        for env in self.iarlenv.rlenv.virtual_peppers:
            env.NO_INERTIA = True

    def step(self, action):
        obs, rew, done, info = self.iarlenv.step(action, ONLY_FOR_AGENT_0=True)
        if self.iarlenv.rlenv.episode_damage > 0:
            done = True
        obs = (obs[0][-1, :, 0], obs[1])  # latest scan only (buffer, ray, channel)
        return obs, rew, done, info

    def render(self, *args, **kwargs):
        self.iarlenv.render(*args, **kwargs)

    def reset(self):
        self.iarlenv.reset()

    def close(self):
        self.iarlenv.close()

    def _get_dt(self):
        return self.iarlenv.args.dt

    def _get_viewer(self):
        viewer = self.iarlenv._get_viewer()
        return viewer


if __name__ == "__main__":
    from navrep.tools.envplayer import EnvPlayer

    env = ToyEnv()
    player = EnvPlayer(env)



import pandas as pd
from matplotlib import pyplot as plt
import os
import sys
import numpy as np

UPDATABLE = True

LOGDIR = os.path.expanduser("~/navrep/logs/W")
if len(sys.argv) == 2:
    LOGDIR = sys.argv[1]

plt.close('all')
while True:
    plt.figure("training log")
    plt.clf()
    logs = sorted(os.listdir(LOGDIR))
    legend = []
    lines = []
    for log in logs:
        if not log.endswith(".csv"):
            continue
        path = os.path.join(LOGDIR, log)
        legend.append(log)
        data = pd.read_csv(path)
        x = data["step"].values
        y = data["cost"].values
        line, = plt.plot(x, y, label=log)
        plt.axhline(np.min(y), alpha=0.3, linewidth=1, color=line.get_color())
        lines.append(line)
    plt.legend()   # quick-search : plot tcn training logs
    plt.ion()
    plt.pause(10.)



from __future__ import print_function
import numpy as np
import os
from sensor_msgs.msg import LaserScan

from navrep.tools.data_extraction import archive_to_lidar_dataset
from navrep.tools.rings import generate_rings
from navrep.models.vae2d import ConvVAE, reset_graph

DEBUG_PLOTTING = True

# Parameters for training
batch_size = 1
NUM_EPOCH = 100
DATA_DIR = "record"
HOME = os.path.expanduser("~")

vae_model_path = os.path.expanduser("~/navrep/models/V/vae.json")

# create network
reset_graph()
vae = ConvVAE(batch_size=batch_size, is_training=False)

# load
vae.load_json(vae_model_path)

# create training dataset
dataset = archive_to_lidar_dataset("~/navrep/datasets/V/ian", limit=180)
if len(dataset) == 0:
    raise ValueError("no scans found, exiting")
print(len(dataset), "scans in dataset.")

# split into batches:
total_length = len(dataset)
num_batches = int(np.floor(total_length / batch_size))

# rings converter
rings_def = generate_rings(64, 64)

dummy_msg = LaserScan()
dummy_msg.range_max = 100.0
dummy_msg.ranges = range(1080)

for idx in range(num_batches):
    batch = dataset[idx * batch_size : (idx + 1) * batch_size]
    scans = batch
    rings = rings_def["lidar_to_rings"](scans).astype(float)

    obs = rings / rings_def["rings_to_bool"]
    # remove "too close" points
    obs[:, :, 0, :] = 0.0

    rings_pred = vae.encode_decode(obs) * rings_def["rings_to_bool"]
    if True:
        import matplotlib.pyplot as plt

        plt.ion()
        plt.figure("rings")
        plt.clf()
        fig, (ax1, ax2) = plt.subplots(
            2, 1, num="rings",
        )
        ax1.imshow(np.reshape(rings, (64, 64)))
        ax2.imshow(np.reshape(rings_pred, (64, 64)))
        ax1.set_title(idx)
        # update
        plt.pause(0.01)



import pandas as pd
from matplotlib import pyplot as plt
import matplotlib
matplotlib.rcParams['mathtext.fontset'] = 'stix'
matplotlib.rcParams['font.family'] = 'STIXGeneral'
import os
import sys
import numpy as np

from navrep.tools.commonargs import parse_plotting_args
from navrep.scripts.plot_gym_training_progress import backend_colors

MAX_GOAL_DIST = 25.

def get_color(log):
    color = None
    if "gpt1d" in log:
        color = backend_colors["GPT1D"]
    elif "gpt" in log:
        color = backend_colors["GPT"]
    elif "vae1dlstm" in log:
        color = backend_colors["VAE1DLSTM"]
    elif "vaelstm" in log:
        color = backend_colors["VAELSTM"]
    elif "navreptrainrnn1d" in log:
        color = backend_colors["VAE1D_LSTM"]
    elif "navreptrainrnn" in log:
        color = backend_colors["VAE_LSTM"]
    return color


if __name__ == "__main__":
    args, _ = parse_plotting_args()

    UPDATABLE = True
    ANIMATE = False

    LOGDIR = args.logdir
    if LOGDIR is None:
        LOGDIR = "~/navrep"
    LOGDIR = os.path.join(LOGDIR, "logs/W")
    LOGDIR = os.path.expanduser(LOGDIR)
    if len(sys.argv) == 2:
        LOGDIR = sys.argv[1]
    M_LOGDIR = LOGDIR.replace("W", "M")

    plt.close('all')
    i_frame = 0
    while True:
        N_ROWS = 2
        N_COLS = 2
        fig, axes = plt.subplots(N_ROWS, N_COLS, num="Worldmodel Error", sharex=True, sharey=False)
        axes = np.array(axes).reshape((N_ROWS, N_COLS))
        plt.cla()
        logs = []
        M_logs = []
        try:
            logs = sorted(os.listdir(LOGDIR))
        except FileNotFoundError:
            pass
        try:
            M_logs = sorted(os.listdir(M_LOGDIR))
        except FileNotFoundError:
            pass
        dirs = [LOGDIR for _ in logs]
        M_dirs = [M_LOGDIR for _ in M_logs]
        legend = []
        lines = []
        for dir_, log in zip((dirs+M_dirs)[::-1], (logs+M_logs)[::-1]):
            if not log.endswith(".csv"):
                continue
            path = os.path.join(dir_, log)
            # get data
            data = pd.read_csv(path)
            if "lidar_test_error" not in data:
                continue
            x = data["step"].values
            y = data["lidar_test_error"].values  # already normalized in wdataset
            y2 = data["state_test_error"].values / MAX_GOAL_DIST**2  # normalize ourself
    #         y = data["cost"].values
            # filter nans
            valid = np.logical_not(np.isnan(y))
            x = x[valid]
            y = y[valid]
            y2 = y2[valid]
            # plot
            if "1d" in log:
                ax = axes[1, 0]
                ax2 = axes[1, 1]
#                 ax.set_title("Lidar State Prediction Error (MSE) - per ray")
#                 ax2.set_title("Goal-Vel State Prediction Error (MSE)")
                ax.set_title("Lidar State Prediction Error (1D)")
                ax2.set_title("Goal-Vel State Prediction Error")
            else:
                ax = axes[0, 0]
                ax2 = axes[0, 1]
#                 ax.set_title("Worldmodel Prediction Error (MSE) - per rings pixel")
#                 ax2.set_title("Goal-Vel State Prediction Error (MSE)")
                ax.set_title("Lidar State Prediction Error")
                ax2.set_title("Goal-Vel State Prediction Error")
            ax.set_yscale('log')
            ax2.set_yscale('log')
            ax.set_ylabel("MSE Error [unitless]")
            ax.set_xlabel("# of training steps")
            color = get_color(log)
            if ANIMATE:
                yanim = np.ones_like(y) * np.nan
                y2anim = np.ones_like(y2) * np.nan
                line, = ax.plot(x, yanim, color=color, label=log)
                line2, = ax2.plot(x, y2anim, color=line.get_c(), linestyle="--")
                line.set_linewidth(2)
                line2.set_linewidth(2)
                N = 10
                for i in range(0, len(y), N):
                    yanim[i:i+N] = y[i:i+N]
                    y2anim[i:i+N] = y2[i:i+N]
                    line.set_ydata(yanim)
                    line2.set_ydata(y2anim)
                    ax.set_xlim([-1000, 240000])
                    ax2.set_xlim([-1000, 240000])
                    ax.set_ylim([0.005, 0.1])
                    ax2.set_ylim([0.0002, 0.2])
                    plt.pause(0.01)
                    plt.savefig("/tmp/plot_W_error_{:05}.png".format(i_frame))
                    i_frame += 1
                line.set_linewidth(1)
                line2.set_linewidth(1)
            else:
                line, = ax.plot(x, y, color=color, label=log)
                ax2.plot(x, y2, color=line.get_c(), linestyle="--")
            ax.axhline(np.min(y), alpha=0.3, linewidth=1, color=line.get_color())
            lines.append(line)
            legend.append(log)
#         for ax in axes.reshape(-1):
#             ax.legend(bbox_to_anchor=(1.05, 1.))   # quick-search : plot tcn training logs
        plt.ion()
        plt.pause(10.)



from navrep.tools.envplayer import EnvPlayer
from navrep.envs.navreptrainencodedenv import NavRepTrainEncodedEnv
from navrep.tools.commonargs import parse_common_args

if __name__ == "__main__":
    args, _ = parse_common_args()

    env = NavRepTrainEncodedEnv(args.backend, args.encoding, scenario='train', gpu=not args.no_gpu)
    player = EnvPlayer(env, render_mode="rings")



from __future__ import print_function
import numpy as np
import os
from tqdm import tqdm
from pose2d import Pose2D, apply_tf

DOWNSAMPLE = None  # 17 (1080 -> 64 rays)
OLD_FORMAT = False

GOAL_REACHED_RADIUS = 0.5


def rosbag_to_lidar_dataset(
    bag_path="~/rosbags/test.bag", lidar_topics=["/combined_scan"]
):
    import rosbag
    bag_path = os.path.expanduser(bag_path)
    bag = rosbag.Bag(bag_path)

    print(bag_path)
    print("------------------------------------------------------------------------")

    scans = []
    for topic, msg, t in bag.read_messages(topics=lidar_topics):
        scans.append(msg.ranges)

    scans = np.array(scans)
    data = raw_scans_to_x(scans)
    return data


def rosbag_to_image_dataset(
    bag_path="~/rosbags/test.bag",
    image_topics=["/camera/color/image_raw"],
    archive_dir="~/navrep/datasets/V/im",
    resize_dim=(64, 64),
):
    import rosbag
    import cv2
    from cv_bridge import CvBridge
    from pyniel.python_tools.path_tools import make_dir_if_not_exists

    archive_dir = os.path.expanduser(archive_dir)

    bridge = CvBridge()

    bag_path = os.path.expanduser(bag_path)
    bag_name = os.path.basename(bag_path).replace(".", "_")
    bag = rosbag.Bag(bag_path)

    print(bag_path)
    print("------------------------------------------------------------------------")

    action_topics = ["/cmd_vel"]
    reward_topics = ["/goal_reached"]  # TODO: this is probably not getting published IRL
    images = []
    actions = []
    rewards = []
    dones = []
    current_action = [0, 0, 0]
    current_reward = -0.01
    goal_reward = 100.0
    for topic, msg, t in bag.read_messages(
        topics=image_topics + action_topics + reward_topics
    ):
        if topic in image_topics:
            cv_image = bridge.imgmsg_to_cv2(msg, desired_encoding="passthrough")
            cv_resized = cv2.resize(cv_image, resize_dim)
            images.append(cv_resized)
            actions.append(current_action)
            rewards.append(current_reward)
            dones.append(0)
        if topic in action_topics:
            current_action = [msg.linear.x, msg.linear.y, msg.angular.z]
        if topic in reward_topics:
            images.append(images[-1])
            actions.append(current_action)
            rewards.append(goal_reward)
            dones.append(1)
            break
    # if goal is not reached, make latest observation final
    if not dones[-1]:
        dones[-1] = 1
    # save
    images = np.array(images)
    actions = np.array(actions)
    rewards = np.array(rewards)
    dones = np.array(dones)
    make_dir_if_not_exists(archive_dir)
    archive_path = os.path.join(
        archive_dir, "{}_images_actions_rewards_dones.npz".format(bag_name)
    )
    np.savez_compressed(
        archive_path, images=images, actions=actions, rewards=rewards, dones=dones
    )
    print("{} images found.".format(len(images)))
    print(archive_path, "written.")

    return images, actions, rewards, dones


def folder_to_lidar_dataset(
    directory="~/autoeval/2020_04_05_17_36", lidar_topics=["/combined_scan"]
):
    import rosbag
    directory = os.path.expanduser(directory)
    # list files
    bagfiles = []
    for dirpath, dirnames, filenames in os.walk(directory):
        for filename in [f for f in filenames if f.endswith(".bag")]:
            bagfiles.append(os.path.join(dirpath, filename))
    print("{} bag files found.".format(len(bagfiles)))
    scans = []
    for bag_path in bagfiles:
        bag = rosbag.Bag(bag_path)
        for topic, msg, t in bag.read_messages(topics=lidar_topics):
            scans.append(msg.ranges)
    print("{} scans found.".format(len(scans)))
    scans = np.array(scans)
    data = raw_scans_to_x(scans)
    return data


def msg_to_lidar_dataset(msg):
    """
    msg: LaserScan

    output: ndarray (n, N, channels), typically 0-100 [meters]
    """
    scans = np.array([msg.ranges])
    data = raw_scans_to_x(scans)
    return data


def x_to_raw_scans(data, msg):
    """
    upsamples if necessary

    data: ndarray (n, N, channels), typically 0-100 [meters]
    msg: LaserScan

    output: ndarray (n, n_rays in msg), typically 0-100 [meters]
    """
    N_RAYS = len(msg.ranges)
    if DOWNSAMPLE is not None:
        data = np.repeat(data, DOWNSAMPLE, axis=1)[:, :N_RAYS, :]
    ranges = np.zeros((len(data), N_RAYS,))
    ranges = data[:, :, 0]
    return ranges


def raw_scans_to_x(scans):
    """
    downsamples if necessary

    raw_scans: ndarray (n, scan_size), typically 0-100 [meters]

    output: ndarray (n, N, channels), typically 0-100 [meters]
    """
    if DOWNSAMPLE is not None:
        scans = scans[:, ::DOWNSAMPLE]
    N_RAYS = scans.shape[1]
    data = np.zeros((len(scans), N_RAYS, 1), dtype=np.float32)
    data[:, :, 0] = scans
    return data


def folder_to_archive(
    directory="~/autoeval",
    archive_dir="~/navrep/datasets/V/ian",
    lidar_topics=["/combined_scan"],
):
    import rosbag
    from pyniel.python_tools.path_tools import make_dir_if_not_exists
    # preproc args
    directory = os.path.expanduser(directory)
    archive_dir = os.path.expanduser(archive_dir)
    # list files
    bagfiles = []
    for dirpath, dirnames, filenames in os.walk(directory):
        for filename in [f for f in filenames if f.endswith(".bag")]:
            bagfiles.append(os.path.join(dirpath, filename))
    print("{} bag files found.".format(len(bagfiles)))
    if OLD_FORMAT:
        for i, bag_path in enumerate(bagfiles):
            scans = []
            bag = rosbag.Bag(bag_path)
            for topic, msg, t in bag.read_messages(topics=lidar_topics):
                scans.append(msg.ranges)
            scans = np.array(scans)
            data = raw_scans_to_x(scans)
            make_dir_if_not_exists(archive_dir)
            archive_path = os.path.join(
                archive_dir, "{:03}_{}rays.npy".format(i, data.shape[1])
            )
            np.save(archive_path, data)
            print(archive_path, "written.")
    # rewards
#     action_topics = ["/cmd_vel"]
    action_topics = ["/pepper_robot/odom"]
    reward_topics = ["/goal_reached"]
    goal_topics = ["/move_base_simple/goal"]
    topics = lidar_topics + action_topics + reward_topics + goal_topics
    for i, bag_path in enumerate(bagfiles):
        scans = []
        robotstates = []
        actions = []
        rewards = []
        dones = []
        current_action = np.array([0, 0, 0])
        current_reward = -0.01
        current_goal = None
        current_goal_frame = None
        goal_reward = 100.0
        print("Loading {}...".format(os.path.basename(bag_path)))
        bag = rosbag.Bag(bag_path)
        try:
            import tf_bag
            bag_transformer = tf_bag.BagTfTransformer(bag)
        except ImportError:
            print("WARNING: Failed to import tf_bag. No goal information will be saved.")
            bag_transformer = None
            current_goal = [np.nan, np.nan]
        if bag.get_message_count(topic_filters=goal_topics) == 0:
            print("WARNING: No goal messages ({}) in rosbag. No goal information will be saved.".format(
                goal_topics))
            bag_transformer = None
            current_goal = [np.nan, np.nan]
        for topic, msg, t in tqdm(bag.read_messages(topics=topics),
                                  total=bag.get_message_count(topic_filters=topics)):
            # update robotstate from current values
            current_goal_in_rob = current_goal
            if current_goal is not None and bag_transformer is not None:
                p2_cg_in_rob = Pose2D(bag_transformer.lookupTransform(
                    "base_footprint", current_goal_frame, msg.header.stamp))
                current_goal_in_rob = apply_tf(current_goal[None,:], p2_cg_in_rob)[0]
            robotstate = None
            if current_goal_in_rob is not None:
                robotstate = np.concatenate([current_goal_in_rob, current_action], axis=0)
            # process messages
            if topic in lidar_topics:
                if current_goal is None:
                    continue
                scan = np.array(msg.ranges)
                scans.append(scan * 1.)
                robotstates.append(robotstate * 1.)
                actions.append(current_action * 1.)
                rewards.append(current_reward * 1.)
                dones.append(0)
            if topic in action_topics:
#                 current_action = [msg.linear.x, msg.linear.y, msg.angular.z]  # if msg is cmd_vel
                current_action = np.array([  # if msg is odometry
                    msg.twist.twist.linear.x, msg.twist.twist.linear.y, msg.twist.twist.angular.z])
            if topic in reward_topics:
                if current_goal is None:
                    continue
                scans.append(scans[-1] * 1.)
                robotstates.append(robotstate * 1.)
                actions.append(current_action * 1.)
                rewards.append(goal_reward * 1.)
                dones.append(1)
                current_goal = None
            if topic in goal_topics:
                if bag_transformer is None:
                    continue
                goal_in_msg = np.array([msg.pose.position.x, msg.pose.position.y])
                p2_msg_in_rob = Pose2D(bag_transformer.lookupTransform(
                    "base_footprint", msg.header.frame_id, msg.header.stamp))
                new_goal = apply_tf(goal_in_msg[None,:], p2_msg_in_rob)[0]
                goal_within_reach = np.linalg.norm(new_goal) < GOAL_REACHED_RADIUS
                # reject new goal which is too close
                if goal_within_reach and current_goal is None:
                    continue
                # goal reached, reset goal
                elif goal_within_reach and current_goal is not None:
                    scans.append(scans[-1] * 1.)
                    robotstates.append(robotstate * 1.)
                    actions.append(current_action * 1.)
                    rewards.append(goal_reward * 1.)
                    dones.append(1)
                    current_goal = None
                    current_goal_frame = None
                # new far away goal -> set goal (episode start)
                elif not goal_within_reach and current_goal is None:
                    current_goal = goal_in_msg
                    current_goal_frame = msg.header.frame_id
                # goal change -> episode change
                elif not goal_within_reach and current_goal is not None:
                    scans.append(scans[-1] * 1.)
                    robotstates.append(robotstate * 1.)
                    actions.append(current_action * 1.)
                    rewards.append(0)
                    dones.append(1)
                    current_goal = goal_in_msg
                    current_goal_frame = msg.header.frame_id
        # if goal is not reached, make latest observation final
        if not dones[-1]:
            dones[-1] = 1
        # save
        scans = np.array(scans)
        robotstates = np.array(robotstates)
        actions = np.array(actions)
        rewards = np.array(rewards)
        dones = np.array(dones)
        make_dir_if_not_exists(archive_dir)
        archive_path = os.path.join(
            archive_dir, "{:03}_scans_robotstates_actions_rewards_dones.npz".format(i)
        )
        np.savez_compressed(
            archive_path, scans=scans, robotstates=robotstates, actions=actions, rewards=rewards, dones=dones
        )
        print(archive_path, "written.")


def archive_to_lidar_dataset(directory="~/navrep/datasets/V/ian", limit=None):
    directory = os.path.expanduser(directory)
    # list files
    files = []
    if OLD_FORMAT:
        for dirpath, dirnames, filenames in os.walk(directory):
            for filename in [f for f in filenames if f.endswith(".npy")]:
                files.append(os.path.join(dirpath, filename))
        arrays = []
        for path in files:
            arrays.append(np.load(path))
        data = np.concatenate(arrays, axis=0)
    else:
        for dirpath, dirnames, filenames in os.walk(directory):
            for filename in [
                f for f in filenames if f.endswith("scans_robotstates_actions_rewards_dones.npz")
            ]:
                files.append(os.path.join(dirpath, filename))
        arrays = []
        if limit is None:
            limit = len(files)
        for path in files[:limit]:
            arrays.append(np.load(path)["scans"])
        data = np.concatenate(arrays, axis=0)
    return data



# This Python file uses the following encoding: utf-8
"""autogenerated by genpy from frame_msgs/PersonTrajectory.msg. Do not edit."""
import sys
python3 = True if sys.hexversion > 0x03000000 else False
import genpy
import struct

import frame_msgs.msg
import geometry_msgs.msg
import genpy

class PersonTrajectory(genpy.Message):
  _md5sum = "f24e0646dee7db1c6a27b5d34153f2f7"
  _type = "frame_msgs/PersonTrajectory"
  _has_header = False #flag to mark the presence of a Header object
  _full_text = """# Message defining the trajectory of a tracked person.
#
# The distinction between track and trajectory is that, depending on the
# implementation of the tracker, a single track (i.e. tracked person) might
# change the trajectory if at some point a new trajectory "fits" that track (person)
# better.
#

uint64                   track_id   # Unique identifier of the tracked person.
PersonTrajectoryEntry[]  trajectory # All states of the last T frames of the most likely trajectory of that tracked person.

================================================================================
MSG: frame_msgs/PersonTrajectoryEntry
# Message defining an entry of a person trajectory.
#

duration    age             # age of the track
bool        is_occluded     # if the track is currently not matched by a detection
uint64      detection_id    # id of the corresponding detection in the current cycle (undefined if occluded)

# The following fields are extracted from the Kalman state x and its covariance C

geometry_msgs/PoseWithCovariance    pose   # pose of the track (z value and orientation might not be set, check if corresponding variance on diagonal is > 99999)
geometry_msgs/TwistWithCovariance   twist  # velocity of the track (z value and rotational velocities might not be set, check if corresponding variance on diagonal is > 99999)

================================================================================
MSG: geometry_msgs/PoseWithCovariance
# This represents a pose in free space with uncertainty.

Pose pose

# Row-major representation of the 6x6 covariance matrix
# The orientation parameters use a fixed-axis representation.
# In order, the parameters are:
# (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
float64[36] covariance

================================================================================
MSG: geometry_msgs/Pose
# A representation of pose in free space, composed of position and orientation. 
Point position
Quaternion orientation

================================================================================
MSG: geometry_msgs/Point
# This contains the position of a point in free space
float64 x
float64 y
float64 z

================================================================================
MSG: geometry_msgs/Quaternion
# This represents an orientation in free space in quaternion form.

float64 x
float64 y
float64 z
float64 w

================================================================================
MSG: geometry_msgs/TwistWithCovariance
# This expresses velocity in free space with uncertainty.

Twist twist

# Row-major representation of the 6x6 covariance matrix
# The orientation parameters use a fixed-axis representation.
# In order, the parameters are:
# (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
float64[36] covariance

================================================================================
MSG: geometry_msgs/Twist
# This expresses velocity in free space broken into its linear and angular parts.
Vector3  linear
Vector3  angular

================================================================================
MSG: geometry_msgs/Vector3
# This represents a vector in free space. 
# It is only meant to represent a direction. Therefore, it does not
# make sense to apply a translation to it (e.g., when applying a 
# generic rigid transformation to a Vector3, tf2 will only apply the
# rotation). If you want your data to be translatable too, use the
# geometry_msgs/Point message instead.

float64 x
float64 y
float64 z"""
  __slots__ = ['track_id','trajectory']
  _slot_types = ['uint64','frame_msgs/PersonTrajectoryEntry[]']

  def __init__(self, *args, **kwds):
    """
    Constructor. Any message fields that are implicitly/explicitly
    set to None will be assigned a default value. The recommend
    use is keyword arguments as this is more robust to future message
    changes.  You cannot mix in-order arguments and keyword arguments.

    The available fields are:
       track_id,trajectory

    :param args: complete set of field values, in .msg order
    :param kwds: use keyword arguments corresponding to message field names
    to set specific fields.
    """
    if args or kwds:
      super(PersonTrajectory, self).__init__(*args, **kwds)
      #message fields cannot be None, assign default values for those that are
      if self.track_id is None:
        self.track_id = 0
      if self.trajectory is None:
        self.trajectory = []
    else:
      self.track_id = 0
      self.trajectory = []

  def _get_types(self):
    """
    internal API method
    """
    return self._slot_types

  def serialize(self, buff):
    """
    serialize message into buffer
    :param buff: buffer, ``StringIO``
    """
    try:
      buff.write(_get_struct_Q().pack(self.track_id))
      length = len(self.trajectory)
      buff.write(_struct_I.pack(length))
      for val1 in self.trajectory:
        _v1 = val1.age
        _x = _v1
        buff.write(_get_struct_2i().pack(_x.secs, _x.nsecs))
        _x = val1
        buff.write(_get_struct_BQ().pack(_x.is_occluded, _x.detection_id))
        _v2 = val1.pose
        _v3 = _v2.pose
        _v4 = _v3.position
        _x = _v4
        buff.write(_get_struct_3d().pack(_x.x, _x.y, _x.z))
        _v5 = _v3.orientation
        _x = _v5
        buff.write(_get_struct_4d().pack(_x.x, _x.y, _x.z, _x.w))
        buff.write(_get_struct_36d().pack(*_v2.covariance))
        _v6 = val1.twist
        _v7 = _v6.twist
        _v8 = _v7.linear
        _x = _v8
        buff.write(_get_struct_3d().pack(_x.x, _x.y, _x.z))
        _v9 = _v7.angular
        _x = _v9
        buff.write(_get_struct_3d().pack(_x.x, _x.y, _x.z))
        buff.write(_get_struct_36d().pack(*_v6.covariance))
    except struct.error as se: self._check_types(struct.error("%s: '%s' when writing '%s'" % (type(se), str(se), str(locals().get('_x', self)))))
    except TypeError as te: self._check_types(ValueError("%s: '%s' when writing '%s'" % (type(te), str(te), str(locals().get('_x', self)))))

  def deserialize(self, str):
    """
    unpack serialized message in str into this message instance
    :param str: byte array of serialized message, ``str``
    """
    try:
      if self.trajectory is None:
        self.trajectory = None
      end = 0
      start = end
      end += 8
      (self.track_id,) = _get_struct_Q().unpack(str[start:end])
      start = end
      end += 4
      (length,) = _struct_I.unpack(str[start:end])
      self.trajectory = []
      for i in range(0, length):
        val1 = frame_msgs.msg.PersonTrajectoryEntry()
        _v10 = val1.age
        _x = _v10
        start = end
        end += 8
        (_x.secs, _x.nsecs,) = _get_struct_2i().unpack(str[start:end])
        _x = val1
        start = end
        end += 9
        (_x.is_occluded, _x.detection_id,) = _get_struct_BQ().unpack(str[start:end])
        val1.is_occluded = bool(val1.is_occluded)
        _v11 = val1.pose
        _v12 = _v11.pose
        _v13 = _v12.position
        _x = _v13
        start = end
        end += 24
        (_x.x, _x.y, _x.z,) = _get_struct_3d().unpack(str[start:end])
        _v14 = _v12.orientation
        _x = _v14
        start = end
        end += 32
        (_x.x, _x.y, _x.z, _x.w,) = _get_struct_4d().unpack(str[start:end])
        start = end
        end += 288
        _v11.covariance = _get_struct_36d().unpack(str[start:end])
        _v15 = val1.twist
        _v16 = _v15.twist
        _v17 = _v16.linear
        _x = _v17
        start = end
        end += 24
        (_x.x, _x.y, _x.z,) = _get_struct_3d().unpack(str[start:end])
        _v18 = _v16.angular
        _x = _v18
        start = end
        end += 24
        (_x.x, _x.y, _x.z,) = _get_struct_3d().unpack(str[start:end])
        start = end
        end += 288
        _v15.covariance = _get_struct_36d().unpack(str[start:end])
        self.trajectory.append(val1)
      return self
    except struct.error as e:
      raise genpy.DeserializationError(e) #most likely buffer underfill


  def serialize_numpy(self, buff, numpy):
    """
    serialize message with numpy array types into buffer
    :param buff: buffer, ``StringIO``
    :param numpy: numpy python module
    """
    try:
      buff.write(_get_struct_Q().pack(self.track_id))
      length = len(self.trajectory)
      buff.write(_struct_I.pack(length))
      for val1 in self.trajectory:
        _v19 = val1.age
        _x = _v19
        buff.write(_get_struct_2i().pack(_x.secs, _x.nsecs))
        _x = val1
        buff.write(_get_struct_BQ().pack(_x.is_occluded, _x.detection_id))
        _v20 = val1.pose
        _v21 = _v20.pose
        _v22 = _v21.position
        _x = _v22
        buff.write(_get_struct_3d().pack(_x.x, _x.y, _x.z))
        _v23 = _v21.orientation
        _x = _v23
        buff.write(_get_struct_4d().pack(_x.x, _x.y, _x.z, _x.w))
        buff.write(_v20.covariance.tostring())
        _v24 = val1.twist
        _v25 = _v24.twist
        _v26 = _v25.linear
        _x = _v26
        buff.write(_get_struct_3d().pack(_x.x, _x.y, _x.z))
        _v27 = _v25.angular
        _x = _v27
        buff.write(_get_struct_3d().pack(_x.x, _x.y, _x.z))
        buff.write(_v24.covariance.tostring())
    except struct.error as se: self._check_types(struct.error("%s: '%s' when writing '%s'" % (type(se), str(se), str(locals().get('_x', self)))))
    except TypeError as te: self._check_types(ValueError("%s: '%s' when writing '%s'" % (type(te), str(te), str(locals().get('_x', self)))))

  def deserialize_numpy(self, str, numpy):
    """
    unpack serialized message in str into this message instance using numpy for array types
    :param str: byte array of serialized message, ``str``
    :param numpy: numpy python module
    """
    try:
      if self.trajectory is None:
        self.trajectory = None
      end = 0
      start = end
      end += 8
      (self.track_id,) = _get_struct_Q().unpack(str[start:end])
      start = end
      end += 4
      (length,) = _struct_I.unpack(str[start:end])
      self.trajectory = []
      for i in range(0, length):
        val1 = frame_msgs.msg.PersonTrajectoryEntry()
        _v28 = val1.age
        _x = _v28
        start = end
        end += 8
        (_x.secs, _x.nsecs,) = _get_struct_2i().unpack(str[start:end])
        _x = val1
        start = end
        end += 9
        (_x.is_occluded, _x.detection_id,) = _get_struct_BQ().unpack(str[start:end])
        val1.is_occluded = bool(val1.is_occluded)
        _v29 = val1.pose
        _v30 = _v29.pose
        _v31 = _v30.position
        _x = _v31
        start = end
        end += 24
        (_x.x, _x.y, _x.z,) = _get_struct_3d().unpack(str[start:end])
        _v32 = _v30.orientation
        _x = _v32
        start = end
        end += 32
        (_x.x, _x.y, _x.z, _x.w,) = _get_struct_4d().unpack(str[start:end])
        start = end
        end += 288
        _v29.covariance = numpy.frombuffer(str[start:end], dtype=numpy.float64, count=36)
        _v33 = val1.twist
        _v34 = _v33.twist
        _v35 = _v34.linear
        _x = _v35
        start = end
        end += 24
        (_x.x, _x.y, _x.z,) = _get_struct_3d().unpack(str[start:end])
        _v36 = _v34.angular
        _x = _v36
        start = end
        end += 24
        (_x.x, _x.y, _x.z,) = _get_struct_3d().unpack(str[start:end])
        start = end
        end += 288
        _v33.covariance = numpy.frombuffer(str[start:end], dtype=numpy.float64, count=36)
        self.trajectory.append(val1)
      return self
    except struct.error as e:
      raise genpy.DeserializationError(e) #most likely buffer underfill

_struct_I = genpy.struct_I
def _get_struct_I():
    global _struct_I
    return _struct_I
_struct_36d = None
def _get_struct_36d():
    global _struct_36d
    if _struct_36d is None:
        _struct_36d = struct.Struct("<36d")
    return _struct_36d
_struct_Q = None
def _get_struct_Q():
    global _struct_Q
    if _struct_Q is None:
        _struct_Q = struct.Struct("<Q")
    return _struct_Q
_struct_BQ = None
def _get_struct_BQ():
    global _struct_BQ
    if _struct_BQ is None:
        _struct_BQ = struct.Struct("<BQ")
    return _struct_BQ
_struct_4d = None
def _get_struct_4d():
    global _struct_4d
    if _struct_4d is None:
        _struct_4d = struct.Struct("<4d")
    return _struct_4d
_struct_2i = None
def _get_struct_2i():
    global _struct_2i
    if _struct_2i is None:
        _struct_2i = struct.Struct("<2i")
    return _struct_2i
_struct_3d = None
def _get_struct_3d():
    global _struct_3d
    if _struct_3d is None:
        _struct_3d = struct.Struct("<3d")
    return _struct_3d



from datetime import datetime
import os

from stable_baselines import PPO2
from stable_baselines.common.vec_env import SubprocVecEnv, DummyVecEnv

from navrep.tools.custom_policy import Custom1DPolicy, ARCH, _C
from navrep.envs.e2eenv import E2E1DNavRepEnv
from navrep.tools.sb_eval_callback import NavrepEvalCallback
from navrep.tools.commonargs import parse_common_args

if __name__ == "__main__":
    args, _ = parse_common_args()

    DIR = os.path.expanduser("~/navrep/models/gym")
    LOGDIR = os.path.expanduser("~/navrep/logs/gym")
    if args.dry_run:
        DIR = "/tmp/navrep/models/gym"
        LOGDIR = "/tmp/navrep/logs/gym"
    START_TIME = datetime.now().strftime("%Y_%m_%d__%H_%M_%S")
    CONTROLLER_ARCH = "_{}_C{}".format(ARCH, _C)
    LOGNAME = "e2e1dnavreptrainenv_" + START_TIME + "_PPO" + "_E2E1D" + CONTROLLER_ARCH
    LOGPATH = os.path.join(LOGDIR, LOGNAME + ".csv")
    MODELPATH = os.path.join(DIR, LOGNAME + "_ckpt")
    MODELPATH2 = os.path.join(DIR, "e2e1dnavreptrainenv_latest_PPO_ckpt")
    if not os.path.exists(DIR):
        os.makedirs(DIR)
    if not os.path.exists(LOGDIR):
        os.makedirs(LOGDIR)

    MILLION = 1000000
    TRAIN_STEPS = args.n
    if TRAIN_STEPS is None:
        TRAIN_STEPS = 60 * MILLION

    N_ENVS = 6
    if args.debug:
        env = DummyVecEnv([lambda: E2E1DNavRepEnv(silent=True, scenario='train')]*N_ENVS)
    else:
        env = SubprocVecEnv([lambda: E2E1DNavRepEnv(silent=True, scenario='train')]*N_ENVS,
                            start_method='spawn')
    eval_env = E2E1DNavRepEnv(silent=True, scenario='train')
    def test_env_fn():  # noqa
        return E2E1DNavRepEnv(silent=True, scenario='test')
    cb = NavrepEvalCallback(eval_env, test_env_fn=test_env_fn,
                            logpath=LOGPATH, savepath=MODELPATH, verbose=1)
    model = PPO2(Custom1DPolicy, env, verbose=0)
    model.learn(total_timesteps=TRAIN_STEPS+1, callback=cb)
    obs = env.reset()

    model.save(MODELPATH)
    model.save(MODELPATH2)
    print("Model '{}' saved".format(MODELPATH))

    del model

    model = PPO2.load(MODELPATH)

    env = E2E1DNavRepEnv(silent=True, scenario='train')
    obs = env.reset()
    for i in range(512):
        action, _states = model.predict(obs, deterministic=True)
        obs, _, done, _ = env.step(action)
        if done:
            env.reset()
#         env.render()



import numpy as np
from tqdm import tqdm
import torch
from torch.utils.data.dataloader import DataLoader

from navrep.tools.wdataset import WorldModelDataset
from navrep.models.vae2d import _64
from navrep.models.rnn import MAX_GOAL_DIST

_Z = 32

def log_loss(y_pred, y_true):
    return y_true * np.log(y_pred) + (1 - y_true) * np.log(1 - y_pred)

def mse(y_pred, y_true):
    return np.mean(np.square(y_pred - y_true))

def rnn_worldmodel_error(rnn, test_dataset_folder, vae):
    sequence_size = rnn.hps.max_seq_len-1
    batch_size = rnn.hps.batch_size
    # load dataset
    seq_loader = WorldModelDataset(test_dataset_folder, sequence_size,
                                   channel_first=False, as_torch_tensors=False, file_limit=64)
    batch_loader = DataLoader(seq_loader, shuffle=False, batch_size=batch_size)
    # iterate over batches
    pbar = tqdm(range(len(batch_loader)*batch_size), total=len(batch_loader)*batch_size).__iter__()
    n_batches = 0
    sum_state_error = 0
    sum_lidar_error = 0
    for x, a, y, x_rs, y_rs, dones in batch_loader:
        # for legacy reasons rnn likes x_rs to be normalized
        x_rs = x_rs / MAX_GOAL_DIST

        # convert img to z
        batch_state = rnn.sess.run(rnn.initial_state)
        # doing the whole batch at once is too mem intensive
        z = np.zeros((batch_size, sequence_size, vae.z_size))  # (b, t, 32)
        for i, single_x in enumerate(x):
            pbar.__next__()
            z[i] = vae.encode(single_x)
        batch_z_rs = np.concatenate([z, x_rs], axis=-1)   # (b, t, 34)
        # add 1 to sequence
        batch_z_rs = np.concatenate([batch_z_rs, np.zeros((batch_size, 1, rnn.hps.seq_width))], axis=1)
        a = np.concatenate([a, np.zeros((batch_size, 1, a.shape[-1]))], axis=1)
        dones = np.concatenate([dones, np.zeros((batch_size, 1))], axis=1)
        # predict
        feed = {
            rnn.batch_z_rs: batch_z_rs,
            rnn.batch_action: a,
            rnn.batch_restart: dones,
            rnn.initial_state: batch_state,
        }
        (logmix, out_mean) = rnn.sess.run([rnn.out_logmix, rnn.out_mean], feed)
        # sample deterministic from mean of MDN
        z_rs_pred = out_mean[(np.arange(len(logmix)),np.argmax(logmix, axis=-1))]
        z_rs_pred = z_rs_pred.reshape(batch_size, sequence_size, rnn.hps.seq_width)
        z_pred = z_rs_pred[:, :, :_Z]
        y_rs_pred = z_rs_pred[:, :, _Z:]
        y_pred_rec = np.zeros((batch_size, sequence_size, _64, _64, 1))
        for i, single_z_pred in enumerate(z_pred):
            y_pred_rec[i] = vae.decode(single_z_pred)

        # cancel the effect of x_rs normalization
        y_rs_pred = y_rs_pred * MAX_GOAL_DIST

        sum_lidar_error += mse(y_pred_rec, y.numpy())  # because binary cross entropy is inf for 0 predictions
        sum_state_error += mse(y_rs_pred, y_rs.numpy())  # mean square error loss
        n_batches += 1
    for _ in pbar:
        pass
    lidar_error = sum_lidar_error / n_batches
    state_error = sum_state_error / n_batches

    return lidar_error, state_error

def vae1d_rnn_worldmodel_error(rnn, test_dataset_folder, vae):
    sequence_size = rnn.hps.max_seq_len-1
    batch_size = rnn.hps.batch_size
    # load dataset
    seq_loader = WorldModelDataset(test_dataset_folder, sequence_size,
                                   lidar_mode="scans",
                                   channel_first=False, as_torch_tensors=False, file_limit=64)
    batch_loader = DataLoader(seq_loader, shuffle=False, batch_size=batch_size)
    # iterate over batches
    pbar = tqdm(range(len(batch_loader)*batch_size), total=len(batch_loader)*batch_size).__iter__()
    n_batches = 0
    sum_state_error = 0
    sum_lidar_error = 0
    for x, a, y, x_rs, y_rs, dones in batch_loader:
        # for legacy reasons rnn likes x_rs to be normalized
        x_rs = x_rs / MAX_GOAL_DIST

        # convert img to z
        batch_state = rnn.sess.run(rnn.initial_state)
        # doing the whole batch at once is too mem intensive
        z = np.zeros((batch_size, sequence_size, vae.z_size))  # (b, t, 32)
        for i, single_x in enumerate(x):
            pbar.__next__()
            z[i] = vae.encode(single_x)
        batch_z_rs = np.concatenate([z, x_rs], axis=-1)   # (b, t, 34)
        # add 1 to sequence
        batch_z_rs = np.concatenate([batch_z_rs, np.zeros((batch_size, 1, rnn.hps.seq_width))], axis=1)
        a = np.concatenate([a, np.zeros((batch_size, 1, a.shape[-1]))], axis=1)
        dones = np.concatenate([dones, np.zeros((batch_size, 1))], axis=1)
        # predict
        feed = {
            rnn.batch_z_rs: batch_z_rs,
            rnn.batch_action: a,
            rnn.batch_restart: dones,
            rnn.initial_state: batch_state,
        }
        (logmix, out_mean) = rnn.sess.run([rnn.out_logmix, rnn.out_mean], feed)
        # sample deterministic from mean of MDN
        z_rs_pred = out_mean[(np.arange(len(logmix)),np.argmax(logmix, axis=-1))]
        z_rs_pred = z_rs_pred.reshape(batch_size, sequence_size, rnn.hps.seq_width)
        z_pred = z_rs_pred[:, :, :_Z]
        y_rs_pred = z_rs_pred[:, :, _Z:]
        y_pred_rec = np.zeros((batch_size, sequence_size, 1080, 1))
        for i, single_z_pred in enumerate(z_pred):
            y_pred_rec[i] = vae.decode(single_z_pred)

        # cancel the effect of x_rs normalization
        y_rs_pred = y_rs_pred * MAX_GOAL_DIST

        sum_lidar_error += mse(y_pred_rec, y.numpy())  # because binary cross entropy is inf for 0 predictions
        sum_state_error += mse(y_rs_pred, y_rs.numpy())  # mean square error loss
        n_batches += 1
    for _ in pbar:
        pass
    lidar_error = sum_lidar_error / n_batches
    state_error = sum_state_error / n_batches

    return lidar_error, state_error

def gpt_worldmodel_error(gpt, test_dataset_folder, device):
    sequence_size = gpt.module.block_size
    batch_size = 128
    # load dataset
    seq_loader = WorldModelDataset(test_dataset_folder, sequence_size,
                                   channel_first=True, as_torch_tensors=True, file_limit=64)
    batch_loader = DataLoader(seq_loader, shuffle=False, batch_size=batch_size)
    # iterate over batches
    batch_loader = tqdm(batch_loader, total=len(batch_loader))
    n_batches = 0
    sum_state_error = 0
    sum_lidar_error = 0
    for x, a, y, x_rs, y_rs, dones in batch_loader:

        # place data on the correct device
        x = x.to(device)
        x_rs = x_rs.to(device)
        a = a.to(device)
        y = y.to(device)
        y_rs = y_rs.to(device)
        dones = dones.to(device)

        y_pred_rec, y_rs_pred, _ = gpt(x, x_rs, a, dones)
        y_pred_rec = y_pred_rec.detach().cpu().numpy()
        y_rs_pred = y_rs_pred.detach().cpu().numpy()

        sum_lidar_error += mse(y_pred_rec, y.cpu().numpy())  # because binary cross entropy is inf for 0
        sum_state_error += mse(y_rs_pred, y_rs.cpu().numpy())  # mean square error loss
        n_batches += 1
    lidar_error = sum_lidar_error / n_batches
    state_error = sum_state_error / n_batches
    return lidar_error, state_error

def gpt1d_worldmodel_error(gpt, test_dataset_folder, device):
    sequence_size = gpt.module.block_size
    batch_size = 128
    # load dataset
    seq_loader = WorldModelDataset(test_dataset_folder, sequence_size,
                                   lidar_mode="scans",
                                   channel_first=True, as_torch_tensors=True, file_limit=64)
    batch_loader = DataLoader(seq_loader, shuffle=False, batch_size=batch_size)
    # iterate over batches
    batch_loader = tqdm(batch_loader, total=len(batch_loader))
    n_batches = 0
    sum_state_error = 0
    sum_lidar_error = 0
    for x, a, y, x_rs, y_rs, dones in batch_loader:

        # place data on the correct device
        x = x.to(device)
        x_rs = x_rs.to(device)
        a = a.to(device)
        y = y.to(device)
        y_rs = y_rs.to(device)
        dones = dones.to(device)

        y_pred_rec, y_rs_pred, _ = gpt(x, x_rs, a, dones)
        y_pred_rec = y_pred_rec.detach().cpu().numpy()
        y_rs_pred = y_rs_pred.detach().cpu().numpy()

        sum_lidar_error += mse(y_pred_rec, y.cpu().numpy())  # because binary cross entropy is inf for 0
        sum_state_error += mse(y_rs_pred, y_rs.cpu().numpy())  # mean square error loss
        n_batches += 1
    lidar_error = sum_lidar_error / n_batches
    state_error = sum_state_error / n_batches
    return lidar_error, state_error

def vae1dlstm_worldmodel_error(vae1dlstm, test_dataset_folder, device):
    sequence_size = 32
    batch_size = 128

    config = vae1dlstm.module.config if hasattr(vae1dlstm, "module") else vae1dlstm.config
    # load dataset
    seq_loader = WorldModelDataset(test_dataset_folder, sequence_size,
                                   lidar_mode="scans",
                                   channel_first=True, as_torch_tensors=True, file_limit=64)
    batch_loader = DataLoader(seq_loader, shuffle=False, batch_size=batch_size)
    # iterate over batches
    batch_loader = tqdm(batch_loader, total=len(batch_loader))
    n_batches = 0
    sum_state_error = 0
    sum_lidar_error = 0
    for x, a, y, x_rs, y_rs, dones in batch_loader:

        # initialize sequence rnn_state
        this_batch_size = x.shape[0]  # last batch can be smaller
        h0 = torch.randn(config.lstm.n_layer, this_batch_size, config.lstm.h_size)
        c0 = torch.randn(config.lstm.n_layer, this_batch_size, config.lstm.h_size)

        # place data on the correct device
        x = x.to(device)
        x_rs = x_rs.to(device)
        a = a.to(device)
        y = y.to(device)
        y_rs = y_rs.to(device)
        dones = dones.to(device)
        h0 = h0.to(device)
        c0 = c0.to(device)

        y_pred_rec, y_rs_pred, _ = vae1dlstm(x, x_rs, a, dones, [h0, c0])
        y_pred_rec = y_pred_rec.detach().cpu().numpy()
        y_rs_pred = y_rs_pred.detach().cpu().numpy()

        sum_lidar_error += mse(y_pred_rec, y.cpu().numpy())  # because binary cross entropy is inf for 0
        sum_state_error += mse(y_rs_pred, y_rs.cpu().numpy())  # mean square error loss
        n_batches += 1
    lidar_error = sum_lidar_error / n_batches
    state_error = sum_state_error / n_batches
    return lidar_error, state_error

def vaelstm_worldmodel_error(vaelstm, test_dataset_folder, device):
    sequence_size = 32
    batch_size = 128

    config = vaelstm.module.config if hasattr(vaelstm, "module") else vaelstm.config
    # load dataset
    seq_loader = WorldModelDataset(test_dataset_folder, sequence_size,
                                   channel_first=True, as_torch_tensors=True, file_limit=64)
    batch_loader = DataLoader(seq_loader, shuffle=False, batch_size=batch_size)
    # iterate over batches
    batch_loader = tqdm(batch_loader, total=len(batch_loader))
    n_batches = 0
    sum_state_error = 0
    sum_lidar_error = 0
    for x, a, y, x_rs, y_rs, dones in batch_loader:

        # initialize sequence rnn_state
        this_batch_size = x.shape[0]  # last batch can be smaller
        h0 = torch.randn(config.lstm.n_layer, this_batch_size, config.lstm.h_size)
        c0 = torch.randn(config.lstm.n_layer, this_batch_size, config.lstm.h_size)

        # place data on the correct device
        x = x.to(device)
        x_rs = x_rs.to(device)
        a = a.to(device)
        y = y.to(device)
        y_rs = y_rs.to(device)
        dones = dones.to(device)
        h0 = h0.to(device)
        c0 = c0.to(device)

        y_pred_rec, y_rs_pred, _ = vaelstm(x, x_rs, a, dones, [h0, c0])
        y_pred_rec = y_pred_rec.detach().cpu().numpy()
        y_rs_pred = y_rs_pred.detach().cpu().numpy()

        sum_lidar_error += mse(y_pred_rec, y.cpu().numpy())  # because binary cross entropy is inf for 0
        sum_state_error += mse(y_rs_pred, y_rs.cpu().numpy())  # mean square error loss
        n_batches += 1
    lidar_error = sum_lidar_error / n_batches
    state_error = sum_state_error / n_batches
    return lidar_error, state_error


if __name__ == "__main__":
    """ Test all worldmodels in logdir """
    import os
    from navrep.tools.commonargs import parse_plotting_args

    from navrep.models.rnn import (reset_graph, default_hps, MDNRNN)
    from navrep.models.vae2d import ConvVAE
    from navrep.models.vae1d import Conv1DVAE
    from navrep.models.gpt import GPT, GPTConfig, load_checkpoint
    from navrep.models.gpt1d import GPT1D
    from navrep.models.vae1dlstm import VAE1DLSTM, VAE1DLSTMConfig
    from navrep.models.vaelstm import VAELSTM, VAELSTMConfig

    _G = 2  # goal dimensions
    _A = 3  # action dimensions
    _RS = 5  # robot state
    _L = 1080  # lidar size
    BLOCK_SIZE = 32  # sequence length (context)

    args, _ = parse_plotting_args()

    gpu = True
    if not gpu:
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"  # disable GPU

    # only implemented for SOADRL
    environment = "navreptrain"
    validation_environment = "navreptrain"
    test_dataset_folder = os.path.expanduser("~/navrep/datasets/V/{}".format(validation_environment))

    UPDATABLE = True

    MODELDIR = args.logdir
    if MODELDIR is None:
        MODELDIR = "~/navrep"
    MODELDIR = os.path.join(MODELDIR, "models")
    MODELDIR = os.path.expanduser(MODELDIR)
    W_MODELDIR = os.path.join(MODELDIR, "W")
    M_MODELDIR = os.path.join(MODELDIR, "M")

    M_models = []
    W_models = []
    try:
        W_models = sorted(os.listdir(W_MODELDIR))[::-1]
    except FileNotFoundError:
        pass
    try:
        M_models = sorted(os.listdir(M_MODELDIR))[::-1]
    except FileNotFoundError:
        pass
    W_dirs = [W_MODELDIR for _ in W_models]
    M_dirs = [M_MODELDIR for _ in M_models]
    for dir_, log in zip(W_dirs+M_dirs, W_models+M_models):
        if log.endswith(".hyperparams.pckl"):
            continue
        path = os.path.join(dir_, log)
        print(log)
        # detect backend type
        if environment+"rnn1d" in log:
            backend = "VAE1D_LSTM"
        elif environment+"rnn" in log:
            backend = "VAE_LSTM"
        elif environment+"gpt1d" in log:
            backend = "GPT1D"
        elif environment+"gpt" in log:
            backend = "GPT"
        elif environment+"vae1dlstm" in log:
            backend = "VAE1DLSTM"
        elif environment+"vaelstm" in log:
            backend = "VAELSTM"
        else:
            print("Backend type for {} not known!".format(log))
            continue

        # Get some param values
        if backend == "GPT":
            from navrep.scripts.train_gpt import _Z, _H
        elif backend == "GPT1D":
            from navrep.scripts.train_gpt1d import _Z, _H
        elif backend == "VAE1DLSTM":
            from navrep.scripts.train_vae1dlstm import _Z, _H
        elif backend == "VAELSTM":
            from navrep.scripts.train_vaelstm import _Z, _H
        elif backend == "VAE_LSTM":
            from navrep.scripts.train_vae import _Z
            from navrep.scripts.train_rnn import _H
        elif backend == "VAE1D_LSTM":
            from navrep.scripts.train_vae1d import _Z
            from navrep.scripts.train_rnn import _H

        # load W / M model
        model = None
        if backend == "VAE_LSTM":
            vae_model_path = os.path.join(MODELDIR,"V", environment+"vae.json")
            reset_graph()
            vae = ConvVAE(z_size=_Z, batch_size=1, is_training=False)
            vae.load_json(vae_model_path)
            hps = default_hps()
            hps = hps._replace(seq_width=_Z+_G, action_width=_A, rnn_size=_H)
            rnn = MDNRNN(hps, gpu_mode=gpu)
            rnn.load_json(path)
        elif backend == "VAE1D_LSTM":
            vae_model_path = os.path.join(MODELDIR,"V", environment+"vae1d.json")
            reset_graph()
            reset_graph()
            vae = Conv1DVAE(z_size=_Z, batch_size=1, is_training=False)
            vae.load_json(vae_model_path)
            hps = default_hps()
            hps = hps._replace(seq_width=_Z+_G, action_width=_A, rnn_size=_H)
            rnn = MDNRNN(hps, gpu_mode=gpu)
            rnn.load_json(path)
        elif backend == "GPT":
            mconf = GPTConfig(BLOCK_SIZE, _H)
            model = GPT(mconf, gpu=gpu)
            load_checkpoint(model, path, gpu=gpu)
            vae = model
            rnn = model
        elif backend == "GPT1D":
            mconf = GPTConfig(BLOCK_SIZE, _H)
            model = GPT1D(mconf, gpu=gpu)
            load_checkpoint(model, path, gpu=gpu)
            vae = model
            rnn = model
        elif backend == "VAELSTM":
            mconf = VAELSTMConfig(_Z, _H)
            model = VAELSTM(mconf, gpu=gpu)
            load_checkpoint(model, path, gpu=gpu)
            vae = model
            rnn = model
        elif backend == "VAE1DLSTM":
            mconf = VAE1DLSTMConfig(_Z, _H)
            model = VAE1DLSTM(mconf, gpu=gpu)
            load_checkpoint(model, path, gpu=gpu)
            vae = model
            rnn = model
        else:
            raise NotImplementedError

        # make torch accept the model if not wrapped in DataParallel
        if model is not None:
            model.module = model

        if backend == "VAE_LSTM":
            lidar_e, state_e = rnn_worldmodel_error(rnn, test_dataset_folder, vae)
        elif backend == "VAE1D_LSTM":
            lidar_e, state_e = vae1d_rnn_worldmodel_error(rnn, test_dataset_folder, vae)
        elif backend == "GPT":
            device = 0
            lidar_e, state_e = gpt_worldmodel_error(rnn, test_dataset_folder, device)
        elif backend == "GPT1D":
            device = 0
            lidar_e, state_e = gpt1d_worldmodel_error(rnn, test_dataset_folder, device)
        elif backend == "VAELSTM":
            device = 0
            lidar_e, state_e = vaelstm_worldmodel_error(rnn, test_dataset_folder, device)
        elif backend == "VAE1DLSTM":
            device = 0
            lidar_e, state_e = vae1dlstm_worldmodel_error(rnn, test_dataset_folder, device)

        print(backend, lidar_e, state_e)

        del rnn
        del vae
        del model



# This Python file uses the following encoding: utf-8
"""autogenerated by genpy from frame_msgs/DetectedPerson.msg. Do not edit."""
import sys
python3 = True if sys.hexversion > 0x03000000 else False
import genpy
import struct

import geometry_msgs.msg

class DetectedPerson(genpy.Message):
  _md5sum = "8ad5b83c32a1fa6e227e3c2da966bc82"
  _type = "frame_msgs/DetectedPerson"
  _has_header = False #flag to mark the presence of a Header object
  _full_text = """# Message describing a detection of a person
#

# Unique id of the detection, monotonically increasing over time
uint64          detection_id

# (Pseudo-)probabilistic value between 0.0 and 1.0 describing the detector's confidence in the detection
float64         confidence

# 3D pose (position + orientation) of the *center* of the detection
# check covariance to see which dimensions are actually set! unset dimensions shall have a covariance > 9999
geometry_msgs/PoseWithCovariance    pose

# height of detected person in 3D (in meters)
float64 height

# 2D bbox position (x and y is the top-left corner, width, height) 
float64 bbox_x
float64 bbox_y
float64 bbox_w
float64 bbox_h

# Sensor modality / detector type, see example constants below. 
# used e.g. later in tracking to check which tracks have been visually confirmed
string          modality            

                                    
string          MODALITY_GENERIC_LASER_2D = laser2d
string          MODALITY_GENERIC_LASER_3D = laser3d
string          MODALITY_GENERIC_MONOCULAR_VISION = mono
string          MODALITY_GENERIC_STEREO_VISION = stereo
string          MODALITY_GENERIC_RGBD = rgbd

# ReID output
float32[] embed_vector

================================================================================
MSG: geometry_msgs/PoseWithCovariance
# This represents a pose in free space with uncertainty.

Pose pose

# Row-major representation of the 6x6 covariance matrix
# The orientation parameters use a fixed-axis representation.
# In order, the parameters are:
# (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
float64[36] covariance

================================================================================
MSG: geometry_msgs/Pose
# A representation of pose in free space, composed of position and orientation. 
Point position
Quaternion orientation

================================================================================
MSG: geometry_msgs/Point
# This contains the position of a point in free space
float64 x
float64 y
float64 z

================================================================================
MSG: geometry_msgs/Quaternion
# This represents an orientation in free space in quaternion form.

float64 x
float64 y
float64 z
float64 w
"""
  # Pseudo-constants
  MODALITY_GENERIC_LASER_2D = 'laser2d'
  MODALITY_GENERIC_LASER_3D = 'laser3d'
  MODALITY_GENERIC_MONOCULAR_VISION = 'mono'
  MODALITY_GENERIC_STEREO_VISION = 'stereo'
  MODALITY_GENERIC_RGBD = 'rgbd'

  __slots__ = ['detection_id','confidence','pose','height','bbox_x','bbox_y','bbox_w','bbox_h','modality','embed_vector']
  _slot_types = ['uint64','float64','geometry_msgs/PoseWithCovariance','float64','float64','float64','float64','float64','string','float32[]']

  def __init__(self, *args, **kwds):
    """
    Constructor. Any message fields that are implicitly/explicitly
    set to None will be assigned a default value. The recommend
    use is keyword arguments as this is more robust to future message
    changes.  You cannot mix in-order arguments and keyword arguments.

    The available fields are:
       detection_id,confidence,pose,height,bbox_x,bbox_y,bbox_w,bbox_h,modality,embed_vector

    :param args: complete set of field values, in .msg order
    :param kwds: use keyword arguments corresponding to message field names
    to set specific fields.
    """
    if args or kwds:
      super(DetectedPerson, self).__init__(*args, **kwds)
      #message fields cannot be None, assign default values for those that are
      if self.detection_id is None:
        self.detection_id = 0
      if self.confidence is None:
        self.confidence = 0.
      if self.pose is None:
        self.pose = geometry_msgs.msg.PoseWithCovariance()
      if self.height is None:
        self.height = 0.
      if self.bbox_x is None:
        self.bbox_x = 0.
      if self.bbox_y is None:
        self.bbox_y = 0.
      if self.bbox_w is None:
        self.bbox_w = 0.
      if self.bbox_h is None:
        self.bbox_h = 0.
      if self.modality is None:
        self.modality = ''
      if self.embed_vector is None:
        self.embed_vector = []
    else:
      self.detection_id = 0
      self.confidence = 0.
      self.pose = geometry_msgs.msg.PoseWithCovariance()
      self.height = 0.
      self.bbox_x = 0.
      self.bbox_y = 0.
      self.bbox_w = 0.
      self.bbox_h = 0.
      self.modality = ''
      self.embed_vector = []

  def _get_types(self):
    """
    internal API method
    """
    return self._slot_types

  def serialize(self, buff):
    """
    serialize message into buffer
    :param buff: buffer, ``StringIO``
    """
    try:
      _x = self
      buff.write(_get_struct_Q8d().pack(_x.detection_id, _x.confidence, _x.pose.pose.position.x, _x.pose.pose.position.y, _x.pose.pose.position.z, _x.pose.pose.orientation.x, _x.pose.pose.orientation.y, _x.pose.pose.orientation.z, _x.pose.pose.orientation.w))
      buff.write(_get_struct_36d().pack(*self.pose.covariance))
      _x = self
      buff.write(_get_struct_5d().pack(_x.height, _x.bbox_x, _x.bbox_y, _x.bbox_w, _x.bbox_h))
      _x = self.modality
      length = len(_x)
      if python3 or type(_x) == unicode:
        _x = _x.encode('utf-8')
        length = len(_x)
      buff.write(struct.pack('<I%ss'%length, length, _x))
      length = len(self.embed_vector)
      buff.write(_struct_I.pack(length))
      pattern = '<%sf'%length
      buff.write(struct.pack(pattern, *self.embed_vector))
    except struct.error as se: self._check_types(struct.error("%s: '%s' when writing '%s'" % (type(se), str(se), str(locals().get('_x', self)))))
    except TypeError as te: self._check_types(ValueError("%s: '%s' when writing '%s'" % (type(te), str(te), str(locals().get('_x', self)))))

  def deserialize(self, str):
    """
    unpack serialized message in str into this message instance
    :param str: byte array of serialized message, ``str``
    """
    try:
      if self.pose is None:
        self.pose = geometry_msgs.msg.PoseWithCovariance()
      end = 0
      _x = self
      start = end
      end += 72
      (_x.detection_id, _x.confidence, _x.pose.pose.position.x, _x.pose.pose.position.y, _x.pose.pose.position.z, _x.pose.pose.orientation.x, _x.pose.pose.orientation.y, _x.pose.pose.orientation.z, _x.pose.pose.orientation.w,) = _get_struct_Q8d().unpack(str[start:end])
      start = end
      end += 288
      self.pose.covariance = _get_struct_36d().unpack(str[start:end])
      _x = self
      start = end
      end += 40
      (_x.height, _x.bbox_x, _x.bbox_y, _x.bbox_w, _x.bbox_h,) = _get_struct_5d().unpack(str[start:end])
      start = end
      end += 4
      (length,) = _struct_I.unpack(str[start:end])
      start = end
      end += length
      if python3:
        self.modality = str[start:end].decode('utf-8')
      else:
        self.modality = str[start:end]
      start = end
      end += 4
      (length,) = _struct_I.unpack(str[start:end])
      pattern = '<%sf'%length
      start = end
      end += struct.calcsize(pattern)
      self.embed_vector = struct.unpack(pattern, str[start:end])
      return self
    except struct.error as e:
      raise genpy.DeserializationError(e) #most likely buffer underfill


  def serialize_numpy(self, buff, numpy):
    """
    serialize message with numpy array types into buffer
    :param buff: buffer, ``StringIO``
    :param numpy: numpy python module
    """
    try:
      _x = self
      buff.write(_get_struct_Q8d().pack(_x.detection_id, _x.confidence, _x.pose.pose.position.x, _x.pose.pose.position.y, _x.pose.pose.position.z, _x.pose.pose.orientation.x, _x.pose.pose.orientation.y, _x.pose.pose.orientation.z, _x.pose.pose.orientation.w))
      buff.write(self.pose.covariance.tostring())
      _x = self
      buff.write(_get_struct_5d().pack(_x.height, _x.bbox_x, _x.bbox_y, _x.bbox_w, _x.bbox_h))
      _x = self.modality
      length = len(_x)
      if python3 or type(_x) == unicode:
        _x = _x.encode('utf-8')
        length = len(_x)
      buff.write(struct.pack('<I%ss'%length, length, _x))
      length = len(self.embed_vector)
      buff.write(_struct_I.pack(length))
      pattern = '<%sf'%length
      buff.write(self.embed_vector.tostring())
    except struct.error as se: self._check_types(struct.error("%s: '%s' when writing '%s'" % (type(se), str(se), str(locals().get('_x', self)))))
    except TypeError as te: self._check_types(ValueError("%s: '%s' when writing '%s'" % (type(te), str(te), str(locals().get('_x', self)))))

  def deserialize_numpy(self, str, numpy):
    """
    unpack serialized message in str into this message instance using numpy for array types
    :param str: byte array of serialized message, ``str``
    :param numpy: numpy python module
    """
    try:
      if self.pose is None:
        self.pose = geometry_msgs.msg.PoseWithCovariance()
      end = 0
      _x = self
      start = end
      end += 72
      (_x.detection_id, _x.confidence, _x.pose.pose.position.x, _x.pose.pose.position.y, _x.pose.pose.position.z, _x.pose.pose.orientation.x, _x.pose.pose.orientation.y, _x.pose.pose.orientation.z, _x.pose.pose.orientation.w,) = _get_struct_Q8d().unpack(str[start:end])
      start = end
      end += 288
      self.pose.covariance = numpy.frombuffer(str[start:end], dtype=numpy.float64, count=36)
      _x = self
      start = end
      end += 40
      (_x.height, _x.bbox_x, _x.bbox_y, _x.bbox_w, _x.bbox_h,) = _get_struct_5d().unpack(str[start:end])
      start = end
      end += 4
      (length,) = _struct_I.unpack(str[start:end])
      start = end
      end += length
      if python3:
        self.modality = str[start:end].decode('utf-8')
      else:
        self.modality = str[start:end]
      start = end
      end += 4
      (length,) = _struct_I.unpack(str[start:end])
      pattern = '<%sf'%length
      start = end
      end += struct.calcsize(pattern)
      self.embed_vector = numpy.frombuffer(str[start:end], dtype=numpy.float32, count=length)
      return self
    except struct.error as e:
      raise genpy.DeserializationError(e) #most likely buffer underfill

_struct_I = genpy.struct_I
def _get_struct_I():
    global _struct_I
    return _struct_I
_struct_5d = None
def _get_struct_5d():
    global _struct_5d
    if _struct_5d is None:
        _struct_5d = struct.Struct("<5d")
    return _struct_5d
_struct_36d = None
def _get_struct_36d():
    global _struct_36d
    if _struct_36d is None:
        _struct_36d = struct.Struct("<36d")
    return _struct_36d
_struct_Q8d = None
def _get_struct_Q8d():
    global _struct_Q8d
    if _struct_Q8d is None:
        _struct_Q8d = struct.Struct("<Q8d")
    return _struct_Q8d
